####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

"""This module implements a partial SPICE netlist parser.

See the :command:`cir2py` tool for an example of usage of the parser.

It would be difficult to implement a full parser for Ngspice since the syntax is mainly contextual.

"""

####################################################################################################

import logging
import os

####################################################################################################

from .BasicElement import SubCircuitElement
from .ElementParameter import FlagParameter
from .Netlist import ElementParameterMetaClass, Circuit, SubCircuit

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class ParseError(NameError):
    pass

####################################################################################################

class PrefixData:

    """This class represents a device prefix."""

    ##############################################

    def __init__(self, prefix, classes):

        self.prefix = prefix
        self.classes = classes

        number_of_positionals_min = 1000
        number_of_positionals_max = 0
        has_optionals = False
        for element_class in classes:
            number_of_positionals = element_class.number_of_positional_parameters
            number_of_positionals_min = min(number_of_positionals_min, number_of_positionals)
            number_of_positionals_max = max(number_of_positionals_max, number_of_positionals)
            has_optionals = max(has_optionals, bool(element_class.optional_parameters))

        self.number_of_positionals_min = number_of_positionals_min
        self.number_of_positionals_max = number_of_positionals_max
        self.has_optionals = has_optionals

        self.multi_devices = len(classes) > 1
        self.has_variable_number_of_pins = prefix in ('Q', 'X') # NPinElement, Q has 3 to 4 pins
        if self.has_variable_number_of_pins:
            self.number_of_pins = None
        else:
            # Q and X are single
            self.number_of_pins = classes[0].number_of_pins

        self.has_flag = False
        for element_class in classes:
            for parameter in element_class.optional_parameters.values():
                if isinstance(parameter, FlagParameter):
                    self.has_flag = True

    ##############################################

    def __len__(self):
        return len(self.classes)

    ##############################################

    def __iter__(self):
        return iter(self.classes)

    ##############################################

    @property
    def single(self):
        if not self.multi_devices:
            return self.classes[0]
        else:
            raise NameError()

####################################################################################################

_prefix_cache = {}
for prefix, classes in ElementParameterMetaClass._classes.items():
    prefix_data = PrefixData(prefix, classes)
    _prefix_cache[prefix] = prefix_data
    _prefix_cache[prefix.lower()] = prefix_data

# for prefix_data in sorted(_prefix_cache.values(), key=lambda x: len(x)):
#     print(prefix_data.prefix,
#           len(prefix_data),
#           prefix_data.number_of_positionals_min, prefix_data.number_of_positionals_max,
#           prefix_data.has_optionals)

# Single:
# B 0 True
# D 1 True
# F 2 False
# G 1 False
# H 2 False
# I 1 False
# J 1 True
# K 3 False
# M 1 True
# S 2 False
# V 1 False
# W 3 False
# Z 1 True

# Two:
# E 0 1 False
# L 1 2 True

# Three:
# C 1 2 True
# R 1 2 True

# NPinElement:
# Q 1 1 True
# X 1 1 False

####################################################################################################

class Statement:

    """ This class implements a statement, in fact a line in a Spice netlist. """

    ##############################################

    def __init__(self, line, statement=None):

        self._line = line

        if statement is not None:
            self._line.lower_case_statement(statement)

    ##############################################

    def __repr__(self):
        return '{} {}'.format(self.__class__.__name__, repr(self._line))

    ##############################################

    def value_to_python(self, x):

        if x:
            if str(x)[0].isdigit():
                return str(x)
            else:
                return "'{}'".format(x)
        else:
            return ''

    ##############################################

    def values_to_python(self, values):

        return [self.value_to_python(x) for x in values]

    ##############################################

    def kwargs_to_python(self, kwargs):
        return ['{}={}'.format(key, self.value_to_python(value))
                for key, value in kwargs.items()]

    ##############################################

    def join_args(self, args):
        return ', '.join(args)

####################################################################################################

class Comment(Statement):
    pass

####################################################################################################

class Title(Statement):

    """ This class implements a title definition. """

    ##############################################

    def __init__(self, line):

        super().__init__(line, statement='title')
        self._title = self._line.right_of('.title')

    ##############################################

    def __str__(self):
        return self._title

    ##############################################

    def __repr__(self):
        return 'Title {}'.format(self._title)

####################################################################################################

class Lib(Statement):

    """ This class implements a library definition. """

    ##############################################

    def __init__(self, line):

        super().__init__(line, statement='lib')
        self._lib = self._line.right_of('.lib')

    ##############################################

    def __str__(self):
        return self._lib

    ##############################################

    def __repr__(self):
        return 'Lib {}'.format(self._lib)

    ##############################################

    def to_python(self, netlist_name):

        return '{}.lib({})'.format(netlist_name, self._lib) + os.linesep

####################################################################################################

class Include(Statement):

    """ This class implements a include definition. """

    ##############################################

    def __init__(self, line):

        super().__init__(line, statement='include')
        self._include = self._line.right_of('.include').strip('"')

    ##############################################

    def __str__(self):
        return self._include

    ##############################################

    def __repr__(self):
        return 'Include {}'.format(self._include)

    ##############################################

    def to_python(self, netlist_name):

        return '{}.include({})'.format(netlist_name, self._include) + os.linesep

####################################################################################################

class Model(Statement):

    """ This class implements a model definition.

    Spice syntax::

        .model mname type (pname1=pval1 pname2=pval2)

    """

    ##############################################

    def __init__(self, line):

        super().__init__(line, statement='model')

        text = line.right_of('.model').strip()
        import re
        mtch = re.match('\s*([^ \t]+)\s*([^ \t(]+)(.*)', text)
        self._name = mtch[1]
        self._model_type = mtch[2]
        params = mtch[3]
        params = params.strip('() ')
        self._parameters = Line.get_kwarg(params)

    ##############################################

    @property
    def name(self):
        """ Name of the model """
        return self._name

    ##############################################

    def __repr__(self):
        return 'Model {} {} {}'.format(self._name, self._model_type, self._parameters)

    ##############################################

    def to_python(self, netlist_name):
        args = self.values_to_python((self._name, self._model_type))
        kwargs = self.kwargs_to_python(self._parameters)
        return '{}.model({})'.format(netlist_name, self.join_args(args + kwargs)) + os.linesep

    ##############################################

    def build(self, circuit):
        circuit.model(self._name, self._model_type, **self._parameters)

####################################################################################################

class SubCircuitStatement(Statement):

    """ This class implements a sub-circuit definition.

    Spice syntax::

        .SUBCKT name node1 ... param1=value1 ...

    """

    ##############################################

    def __init__(self, line):

        super().__init__(line, statement='subckt')

        # Fixme
        parameters, dict_parameters = self._line.split_line('.subckt')
        self._name, self._nodes = parameters[0], parameters[1:]

        self._statements = []

    ##############################################

    @property
    def name(self):
        """ Name of the sub-circuit. """
        return self._name

    @property
    def nodes(self):
        """ Nodes of the sub-circuit. """
        return self._nodes

    ##############################################

    def __repr__(self):
        text = 'SubCircuit {} {}'.format(self._name, self._nodes) + os.linesep
        text += os.linesep.join(['  ' + repr(statement) for statement in self._statements])
        return text

    ##############################################

    def __iter__(self):
        """ Return an iterator on the statements. """
        return iter(self._statements)

    ##############################################

    def append(self, statement):
        """ Append a statement to the statement's list. """
        self._statements.append(statement)

    ##############################################

    def to_python(self, ground=0):

        subcircuit_name = 'subcircuit_' + self._name
        args = self.values_to_python([subcircuit_name] + self._nodes)
        source_code = ''
        source_code += '{} = SubCircuit({})'.format(subcircuit_name, self.join_args(args)) + os.linesep
        source_code += SpiceParser.netlist_to_python(subcircuit_name, self, ground)
        return source_code

    ##############################################

    def build(self, ground=0):
        subcircuit = SubCircuit(self._name, *self._nodes)
        SpiceParser._build_circuit(subcircuit, self._statements, ground)
        return subcircuit

####################################################################################################

class Element(Statement):

    """ This class implements an element definition.

    "{ expression }" are allowed in device line.

    """

    _logger = _module_logger.getChild('Element')

    ##############################################

    def __init__(self, line):

        super().__init__(line)

        line_str = str(line)
        # self._logger.debug(os.linesep + line_str)

        # Retrieve device prefix
        self._prefix = line_str[0]
        prefix_data = _prefix_cache[self._prefix]

        # Retrieve device name
        start_location = 1
        stop_location = line_str.find(' ')
        # Fixme: if stop_location == -1:
        self._name = line_str[start_location:stop_location]

        self._nodes = []
        self._parameters = []
        self._dict_parameters = {}

        # Read nodes
        if not prefix_data.has_variable_number_of_pins:
            number_of_pins = prefix_data.number_of_pins
            if number_of_pins:
                self._nodes, stop_location = self._line.read_words(stop_location, number_of_pins)
        else: # Q or X
            if prefix_data.prefix == 'Q':
                self._nodes, stop_location = self._line.read_words(stop_location, 3)
                # Fixme: optional node
            else: # X
                args, stop_location = self._line.split_words(stop_location, until='=')
                self._nodes = args[:-1]
                self._parameters.append(args[-1]) # model name

        # Read positionals
        number_of_positionals = prefix_data.number_of_positionals_min
        if number_of_positionals and stop_location is not None: # model is optional
            self._parameters, stop_location = self._line.read_words(stop_location, number_of_positionals)
        if prefix_data.multi_devices and stop_location is not None:
            remaining, stop_location = self._line.split_words(stop_location, until='=')
            self._parameters.extend(remaining)

        if prefix_data.prefix in ('V', 'I') and stop_location is not None:
            # merge remaining
            self._parameters[-1] += line_str[stop_location:]

        # Read optionals
        if prefix_data.has_optionals and stop_location is not None:
            kwargs, stop_location = self._line.split_words(stop_location)
            for kwarg in kwargs:
                try:
                    key, value = kwarg.split('=')
                    self._dict_parameters[key.lower()] = value
                except ValueError:
                    if kwarg in ('off',) and prefix_data.has_flag:
                        self._dict_parameters['off'] = True
                    else:
                        # Fixme: warning -> debug due to spam ...
                        self._logger.debug(line_str)
                        # raise NameError('Bad element line:', line_str)

        if prefix_data.multi_devices:
            for element_class in prefix_data:
                if len(self._parameters) == element_class.number_of_positional_parameters:
                    break
        else:
            element_class = prefix_data.single
        self.factory = element_class

        # Move positionals passed as kwarg
        to_delete = []
        for parameter in element_class.positional_parameters.values():
            if parameter.key_parameter:
                i = parameter.position
                self._dict_parameters[parameter.attribute_name] = self._parameters[i]
                to_delete.append(i)
        for i in to_delete:
            del self._parameters[i]

        # self._logger.debug(os.linesep + self.__repr__())

    ##############################################

    @property
    def name(self):
        """ Name of the element """
        return self._name

    ##############################################

    def __repr__(self):
        return 'Element {0._prefix} {0._name} {0._nodes} {0._parameters} {0._dict_parameters}'.format(self)

    ##############################################

    def translate_ground_node(self, ground):

        nodes = []
        for node in self._nodes:
            if str(node) == str(ground):
                node = 0
            nodes.append(node)

        return nodes

    ##############################################

    def to_python(self, netlist_name, ground=0):

        nodes = self.translate_ground_node(ground)
        args = [self._name]
        if self._prefix != 'X':
            args += nodes + self._parameters
        else: # != Spice
            args += self._parameters + nodes
        args = self.values_to_python(args)
        kwargs = self.kwargs_to_python(self._dict_parameters)
        return '{}.{}({})'.format(netlist_name, self._prefix, self.join_args(args + kwargs)) + os.linesep

    ##############################################

    def build(self, circuit, ground=0):

        factory = getattr(circuit, self.factory.ALIAS)
        nodes = self.translate_ground_node(ground)
        if self._prefix != 'X':
            args = nodes + self._parameters
        else: # != Spice
            args = self._parameters + nodes
        kwargs = self._dict_parameters
        if self._logger.isEnabledFor(logging.DEBUG):
            message = ' '.join([str(x) for x in (self._prefix, self._name, nodes,
                                                 self._parameters, self._dict_parameters)])
            self._logger.debug(message)
        factory(self._name, *args, **kwargs)

####################################################################################################

class Line:

    """ This class implements a line in the netlist. """

    _logger = _module_logger.getChild('Element')

    ##############################################

    def __init__(self, line, line_range, end_of_line_comment):

        self._end_of_line_comment = end_of_line_comment

        text, comment, self._is_comment = self._split_comment(line)

        self._text = text
        self._comment = comment
        self._line_range = line_range

    ##############################################

    def __repr__(self):
        return '{0._line_range}: {0._text} // {0._comment}'.format(self)

    ##############################################

    def __str__(self):
        return self._text

    ##############################################

    @property
    def comment(self):
        return self._comment

    @property
    def is_comment(self):
        return self._is_comment

    ##############################################

    def _split_comment(self, line):

        line = str(line)

        if line.startswith('*'):
            is_comment = True
            text = ''
            comment = line[1:].strip()
        else:
            is_comment = False
            # remove end of line comment
            location = -1
            for marker in self._end_of_line_comment:
                _location = line.find(marker)
                if _location != -1:
                    if location == -1:
                        location = _location
                    else:
                        location = min(_location, location)
            if location != -1:
                text = line[:location].strip()
                comment = line[location:].strip()
            else:
                text = line
                comment = ''

        return text, comment, is_comment

    ##############################################

    def append(self, line):

        text, comment, is_comment = self._split_comment(line)

        if text:
            if not self._text.endswith(' ') or text.startswith(' '):
                self._text += ' '
            self._text += text
        if comment:
            self._comment += ' // ' + comment

        _slice = self._line_range
        self._line_range = slice(_slice.start, _slice.stop + 1)

    ##############################################

    def lower_case_statement(self, statement):

        """Lower case the statement"""

        # statement without . prefix

        if self._text:
            lower_statement = statement.lower()
            _slice = slice(1, len(statement) + 1)
            _statement = self._text[_slice]
            if _statement.lower() == lower_statement:
                self._text = '.' + lower_statement + self._text[_slice.stop:]

    ##############################################

    def right_of(self, text):
        return self._text[len(text):].strip()

    ##############################################

    def read_words(self, start_location, number_of_words):

        """Read a fixed number of words separated by space."""

        words = []
        stop_location = None

        line_str = self._text
        number_of_words_read = 0
        while number_of_words_read < number_of_words: # and start_location < len(line_str)
            stop_location = line_str.find(' ', start_location)
            if stop_location == -1:
                stop_location = None # read until end
            word = line_str[start_location:stop_location].strip()
            if word:
                number_of_words_read += 1
                words.append(word)
            if stop_location is None: # we should stop
                if number_of_words_read != number_of_words:
                    template = 'Bad element line, looking for word {}/{}:' + os.linesep
                    message = (template.format(number_of_words_read, number_of_words) +
                               line_str + os.linesep +
                               ' '*start_location + '^')
                    self._logger.warning(message)
                    raise ParseError(message)
            else:
                if start_location < stop_location:
                    start_location = stop_location
                else: # we have read a space
                    start_location += 1

        return words, stop_location

    ##############################################

    def split_words(self, start_location, until=None):

        stop_location = None

        line_str = self._text
        if until is not None:
            location = line_str.find(until, start_location)
            if location != -1:
                stop_location = location
                location = line_str.rfind(' ', start_location, stop_location)
                if location != -1:
                    stop_location = location
                else:
                    raise NameError('Bad element line, missing key? ' + line_str)

        line_str = line_str[start_location:stop_location]
        words = [x for x in line_str.split(' ') if x]

        return words, stop_location

    ##############################################

    @staticmethod
    def get_kwarg(text):

        dict_parameters = {}

        parts = []
        for part in text.split():
            if '=' in part and part != '=':
                left, right = [x for x in part.split('=')]
                parts.append(left)
                parts.append('=')
                if right:
                    parts.append(right)
            else:
                parts.append(part)

        i = 0
        i_stop = len(parts)
        while i < i_stop:
            if i + 1 < i_stop and parts[i + 1] == '=':
                key, value = parts[i], parts[i + 2]
                dict_parameters[key] = value
                i += 3
            else:
                raise ParseError("Bad kwarg: {}".format(text))

        return dict_parameters

    ##############################################

    def split_line(self, keyword):

        """Split the line according to the following pattern::

            keyword parameter1 parameter2 ... key1=value1 key2=value2 ...

        Return the list of parameters and the dictionary.

        """

        # Fixme: cf. get_kwarg

        parameters = []
        dict_parameters = {}

        text = self.right_of(keyword)

        parts = []
        for part in text.split():
            if '=' in part and part != '=':
                left, right = [x for x in part.split('=')]
                parts.append(left)
                parts.append('=')
                if right:
                    parts.append(right)
            else:
                parts.append(part)

        i = 0
        i_stop = len(parts)
        while i < i_stop:
            if i + 1 < i_stop and parts[i + 1] == '=':
                key, value = parts[i], parts[i + 2]
                dict_parameters[key] = value
                i += 3
            else:
                parameters.append(parts[i])
                i += 1

        return parameters, dict_parameters

####################################################################################################

class SpiceParser:

    """ This class parse a Spice netlist file and build a syntax tree.

    Public Attributes:

      :attr:`circuit`

      :attr:`models`

      :attr:`subcircuits`

      :attr:`incl_libs`

    """

    _logger = _module_logger.getChild('SpiceParser')

    ##############################################

    def __init__(self, path=None, source=None, end_of_line_comment=('$', '//', ';'), recurse=False, section=None):

        # Fixme: empty source

        self._path = path  # For use by _parse() when recursing through files.

        if path is not None:
            with open(str(path), 'r') as f:
                raw_lines = f.readlines()
        elif source is not None:
            raw_lines = source.split(os.linesep)
        else:
            raise ValueError

        self._end_of_line_comment = end_of_line_comment

        lines = self._merge_lines(raw_lines)
        self._title = None
        self._statements = self._parse(lines=lines, recurse=recurse, section=section)
        self._find_sections()

    ##############################################

    def _merge_lines(self, raw_lines):

        """Merge broken lines and return a new list of lines.

        A line starting with "+" continues the preceding line.
        """

        lines = []
        current_line = None
        for line_index, line_string in enumerate(raw_lines):
            line_string = line_string.lstrip(' ')
            if line_string.startswith('+'):
                current_line.append(line_string[1:].strip('\r\n'))
            else:
                line_string = line_string.strip('\r\n')
                if line_string:
                    _slice = slice(line_index, line_index +1)
                    line = Line(line_string, _slice, self._end_of_line_comment)
                    lines.append(line)
                    # handle case with comment before line continuation
                    if not line_string.startswith('*'):
                        current_line = line

        return lines

    ##############################################

    def _parse(self, lines, recurse=False, section=None):

        """ Parse the lines and return a list of statements. """

        # The first line in the input file must be the title, which is the only comment line that does
        # not need any special character in the first place.
        #
        # The last line must be .end

        if len(lines) <= 1:
            self._logger.warning('Empty Spice file: {self._path}'.format(**locals()))
            # raise NameError('Netlist is empty')
        # if lines[-1] != '.end':
        #     raise NameError('".end" is expected at the end of the netlist')

        title_statement = '.title '
        self._title = str(lines[0])
        if self._title.startswith(title_statement):
            self._title = self._title[len(title_statement):]

        # SUBCKT and MODEL files often start with their commands as the
        # first line so they'll parse incorrectly if that line is removed.
        # For everything else, assume the first line is a TITLE line and
        # remove it.
        if str(lines[0]).startswith(('.model', '.subckt')):
            start_index = 0
        else:
            start_index = 1

        statements = []
        skip_lines = [False]  # True on top of stack means skip lines.
        sub_circuit = None
        scope = statements
        self.incl_libs = []  # Libraries found during recursive descent into includes.
        for line in lines[start_index:]:
            # print('>', repr(line))
            text = str(line)
            lower_case_text = text.lower() # !
            if skip_lines[-1]:
                if lower_case_text.startswith('.endl'):
                    skip_lines.pop()
            elif line.is_comment:
                scope.append(Comment(line))
            elif lower_case_text.startswith('.'):
                lower_case_text = lower_case_text[1:]
                if lower_case_text.startswith('subckt'):
                    sub_circuit = SubCircuitStatement(line)
                    statements.append(sub_circuit)
                    scope = sub_circuit
                elif lower_case_text.startswith('ends'):
                    sub_circuit = None
                    scope = statements
                elif lower_case_text.startswith('title'):
                    # override first line
                    self._title = Title(line)
                    scope.append(self._title)
                elif lower_case_text.startswith('end'):
                    pass
                elif lower_case_text.startswith('model'):
                    model = Model(line)
                    scope.append(model)
                elif lower_case_text.startswith('include'):
                    incl = Include(line)
                    scope.append(incl)
                    if recurse:
                        from .Library import SpiceLibrary
                        incl_path = os.path.join(str(self._path.directory_part()), str(incl))
                        self.incl_libs.append(SpiceLibrary(root_path=incl_path, recurse=recurse))
                elif lower_case_text.startswith('lib'):
                    lib = Lib(line)
                    if section and str(lib) != section.lower():
                        # If the .lib statement is only followed by the name of a section,
                        # then skip any lines in a library section whose name does not match
                        # the library section argument.
                        skip_lines.append(True)
                    else:
                        scope.append(lib)
                else:
                    # options param ...
                    # .global
                    # .lib filename libname
                    # .param
                    # .func .csparam .temp .if
                    # { expr } are allowed in .model lines and in device lines.
                    # self._logger.warning('Parser ignored: {}'.format(line))
                    pass
            else:
                try:
                    element = Element(line)
                    scope.append(element)
                except ParseError:
                    self._logger.warning('Parse error on:\n{}'.format(line))

        return statements

    ##############################################

    def _find_sections(self):

        """ Look for model, sub-circuit and circuit definitions in the statement list. """

        self.circuit = None
        self.subcircuits = []
        self.models = []
        for statement in self._statements:
            if isinstance(statement, Title):
                if self.circuit is None:
                    self.circuit = statement
                else:
                    raise NameError('More than one title')
            elif isinstance(statement, SubCircuitStatement):
                self.subcircuits.append(statement)
            elif isinstance(statement, Model):
                self.models.append(statement)

    ##############################################

    def is_only_subcircuit(self):
        return bool(not self.circuit and self.subcircuits)

    ##############################################

    def is_only_model(self):
        return bool(not self.circuit and not self.subcircuits and self.models)

    ##############################################

    @staticmethod
    def _build_circuit(circuit, statements, ground):

        for statement in statements:
            if isinstance(statement, Include):
                circuit.include(str(statement))

        for statement in statements:
            if isinstance(statement, Element):
                statement.build(circuit, ground)
            elif isinstance(statement, Model):
                statement.build(circuit)
            elif isinstance(statement, SubCircuit):
                subcircuit = statement.build(ground) # Fixme: ok ???
                circuit.subcircuit(subcircuit)

    ##############################################

    def build_circuit(self, ground=0):

        """Build a :class:`Circuit` instance.

        Use the *ground* parameter to specify the node which must be translated to 0 (SPICE ground node).

        """

        circuit = Circuit(str(self._title))
        self._build_circuit(circuit, self._statements, ground)
        return circuit

    ##############################################

    @staticmethod
    def netlist_to_python(netlist_name, statements, ground=0):

        source_code = ''
        for statement in statements:
            if isinstance(statement, Element):
                source_code += statement.to_python(netlist_name, ground)
            elif isinstance(statement, Include):
                pass
            elif isinstance(statement, Model):
                source_code += statement.to_python(netlist_name)
            elif isinstance(statement, SubCircuitStatement):
                source_code += statement.to_python(netlist_name)
            elif isinstance(statement, Include):
                source_code += statement.to_python(netlist_name)
        return source_code

    ##############################################

    def to_python_code(self, ground=0):

        ground = str(ground)

        source_code = ''

        if self.circuit:
            source_code += "circuit = Circuit('{}')".format(self._title) + os.linesep
        source_code += self.netlist_to_python('circuit', self._statements, ground)

        return source_code



####################################################################################################

from PySpice.Spice.Netlist import SubCircuitFactory
from PySpice.Unit import *

####################################################################################################

class Transformer(SubCircuitFactory):

    NAME = 'Transformer'
    NODES = ('input_plus', 'input_minus',
             'output_plus', 'output_minus')

    ##############################################

    def __init__(self,
                 turn_ratio,
                 primary_inductance=1@u_H,
                 copper_resistance=1@u_Ω,
                 leakage_inductance=1@u_mH,
                 winding_capacitance=20@u_pF,
                 coupling=.999,
             ):

        super().__init__()

        # For an ideal transformer you can reduce the values for the flux leakage inductances, the
        # copper resistors and the winding capacitances. But
        if copper_resistance <= 0:
            raise ValueError("copper resistance must be > 0")
        if leakage_inductance <= 0:
            raise ValueError("leakage inductance must be > 0")

        # primary_turns =
        # secondary_turns =
        # turn_ratio = primary_turns / secondary_turns
        # primary_inductance =
        # primary_inductance / secondary_inductance = turn_ratio**2
        secondary_inductance = primary_inductance / float(turn_ratio**2)

        # Primary
        self.C('primary', 'input_plus', 'input_minus', winding_capacitance)
        self.L('primary_leakage', 'input_plus', 1, leakage_inductance)
        primary_inductor = self.L('primary', 1, 2, primary_inductance)
        self.R('primary', 2, 'output_minus', copper_resistance)

        # Secondary
        self.C('secondary', 'output_plus', 'output_minus', winding_capacitance)
        self.L('secondary_leakage', 'output_plus', 3, leakage_inductance)
        secondary_inductor = self.L('secondary', 3, 4, secondary_inductance)
        self.R('secondary', 4, 'output_minus', copper_resistance)

        # Coupling
        self.CoupledInductor('coupling', primary_inductor.name, secondary_inductor.name, coupling)



####################################################################################################
#
# PySpice - A Spice package for Python
# Copyright (C) 2019 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

from pathlib import Path
import os

from invoke import task

####################################################################################################

def find(matcher, path='.'):
    to_delete = []
    for root, _, filenames in os.walk('.'):
        root = Path(root)
        for filename in filenames:
            filename = Path(filename)
            if matcher(filename):
                path = root.joinpath(filename)
                to_delete.append(path)
    if to_delete:
        rule = '='*100
        print(rule)
        for path in to_delete:
            print(path)
        print(rule)
        rc = input('remove ? [n]/y ')
        if rc == 'y':
            for path in to_delete:
                path.unlink(missing_ok=True)

####################################################################################################

@task
def flycheck(ctx):
    with ctx.cd(ctx.Package):
        # ctx.run('find . -name "flycheck*.py" -exec /usr/bin/rm {} \;')
        find(lambda filename: filename.suffix == '.py' and filename.stem.startswith('flycheck'))

@task
def emacs_backup(ctx):
    # ctx.run('find . -name "*~" -type f -exec /usr/bin/rm -f {} \;')
    find(lambda filename: str(filename).endswith('~'))

@task(flycheck, emacs_backup)
def clean(ctx):
    pass



####################################################################################################
#
# PySpice - A Spice package for Python
# Copyright (C) 2020 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

from pathlib import Path
import os
import subprocess
import sys
import tempfile

from invoke import task

####################################################################################################

_ = sys.platform
is_linux = _.startswith('linux')
is_osx = _.startswith('darwin')
is_windows = _.startswith('win')
if not (is_linux or is_osx or is_windows):
    raise NotImplementedError

####################################################################################################

on_azure = os.environ.get('CI', None) == 'azure'

####################################################################################################

if is_windows:
    # Windows uses CP1252 encoding for io
    # $env:PYTHONIOENCODING="utf_8"
    import locale
    print(sys.getdefaultencoding())
    print(locale.getpreferredencoding())

####################################################################################################

PYSPICE_SOURCE_PATH = Path(__file__).resolve().parents[1]
EXAMPLES_PATH = PYSPICE_SOURCE_PATH.joinpath('examples')

####################################################################################################

def make_path(*args):
    return str(Path(EXAMPLES_PATH, *args))

####################################################################################################

def is_example(root, filename):
    if filename.suffix == '.py' and str(filename).islower():
        path = root.joinpath(filename)
        if path.is_symlink():
            return None
        with open(path) as fh:
            for line in fh.readlines()[:2]:
                line = line.strip()
                if line == '#skip#':
                    return None
        return path
    return None

####################################################################################################

def example_iter():

    for root, _, filenames in os.walk(EXAMPLES_PATH):
        root = Path(root)
        for filename in filenames:
            filename = Path(filename)
            path = is_example(root, filename)
            if path is not None:
                # Windows: check file is not a Git symlink
                with open(path) as fh:
                    content = fh.readlines()
                if not (len(content) == 1 and content[0].strip().endswith('.py')):
                    yield path

####################################################################################################

def run_example(path):

    # Windows: tempfile must be closed else
    # c:\python38\python.exe: can't open file
    #  'C:\Users\travis\build\FabriceSalvaire\PySpice\examples\...\tmpgx3uzlad.py':
    #  [Errno 13] Permission denied

    # Comment plt.show()
    #  annoying on a Linux terminal but works on Travis
    #  hang on OSX
    #  Windows ???
    with tempfile.NamedTemporaryFile(dir=path.parent, suffix='.py', delete=False) as tmp_fh:
        tmp_path = Path(tmp_fh.name)
        with open(path, encoding='utf-8') as fh:
            content = fh.read()
        content = content.replace('plt.show()', '#plt.show()')
        tmp_fh.write(content.encode('utf-8'))

    print('Run {}'.format(path))
    # _ = path
    _ = tmp_path
    command = (sys.executable, str(_)) # else TypeError: argument of type 'WindowsPath' is not iterable
    process = subprocess.run(command, capture_output=True, encoding='utf-8')
    tmp_path.unlink()
    if process.returncode:
        print(process.stdout)
        print(process.stderr)
        return False
    else:
        return True

####################################################################################################

def on_linux(path):

    skipped_files = [
        make_path('operational-amplifier', 'astable.py'),   # doAnalyses: Too many iterations without convergence
    ]

    if str(path) in skipped_files:
        print('Skip {}'.format(path))
        return 'skipped'

    return run_example(path)

####################################################################################################

def on_osx(path):

    skipped_files = [
        make_path('ngspice-shared', 'external-source.py'),
    ]

    if on_azure:
        skipped_files += [
            # Error: ngspice.dll cannot recover and awaits to be detached
            make_path('ngspice-shared', 'ngspice-interpreter.py'),
        ]

    if str(path) in skipped_files:
        print('Skip {}'.format(path))
        return 'skipped'

    return run_example(path)

####################################################################################################

def on_windows(path):

    skipped_files = []

    print('on_azure', on_azure, os.environ.get('CI', None))
    # if on_azure:
    # %SRC_DIR%\examples\switched-power-supplies\buck-converter.py
    skipped_files += [
        make_path('basic-usages', 'unit.py'),
        make_path('switched-power-supplies', 'buck-converter.py'),
        make_path('operational-amplifier', 'astable.py'),   # doAnalyses: Too many iterations without convergence
    ]

    if str(path) in skipped_files:
        print('Skip {}'.format(path))
        return 'skipped'

    return run_example(path)

####################################################################################################

@task()
def run_examples(ctx):

    # subprocess capture_output requires 3.7
    if sys.version_info.minor < 7:
        print('WARNING: Skip tests because Python < 3.7')
        return

    # os.environ['PySpiceLibraryPath'] = str(EXAMPLES_PATH.joinpath('libraries'))

    succeed = []
    failed = []
    skipped = []

    examples = sorted(example_iter(), key=lambda _: str(_))
    for path in examples:
        if is_linux:
            rc = on_linux(path)
        elif is_osx:
            rc = on_osx(path)
        elif is_windows:
            rc = on_windows(path)

        if rc == 'skipped':
            skipped.append(path)
        elif rc:
            succeed.append(path)
        else:
            failed.append(path)

    print()
    for path in skipped:
        print('SKIPPED', path)
    print()
    for path in succeed:
        print('SUCCEED', path)
    print()
    for path in failed:
        print('FAILED', path)
    print()

    if failed:
        raise NameError('Some tests failed')



####################################################################################################
#
# PySpice - A Spice package for Python
# Copyright (C) 2021 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

from invoke import task

try:
    from github import Github
except ImportError:
    pass

####################################################################################################

REPOSITORY_NAME = "FabriceSalvaire/PySpice"

####################################################################################################

def get_repo():
    g = Github()
    repo = g.get_repo(REPOSITORY_NAME)
    return repo

####################################################################################################

@task
def labels(ctx):
    repo = get_repo()
    labels = repo.get_labels()
    for label in labels:
        print(f'{label.name}: {label.description}')



####################################################################################################
#
# PySpice - A Spice package for Python
# Copyright (C) 2019 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

from pathlib import Path
import os
import shutil
import sys

from invoke import task

####################################################################################################

import PySpice.Spice.NgSpice as NgSpice

####################################################################################################

PYSPICE_SOURCE_PATH = Path(__file__).resolve().parents[1]

BASE_URL = 'https://sourceforge.net/projects/ngspice/files'
RELEASE_URL = BASE_URL + '/ng-spice-rework'
RELEASE_NOTE_URL = RELEASE_URL + '/{}/ReleaseNotes.txt/download'
MANUAL_URL = RELEASE_URL + '/{0}/ngspice-{0}-manual.pdf/download'
# LATEST_URL = BASE_URL + '/latest/download' # zip
TAR_URL = RELEASE_URL + '/{0}/ngspice-{0}.tar.gz'
# OSX_URL = RELEASE_URL + '/{0}/ngspice-{0}.pkg'
WINDOWS_URL = RELEASE_URL + '/{0}/ngspice-{0}_64.zip'
WINDOWS_DLL_URL = RELEASE_URL + '/{0}/ngspice-{0}_dll_64.zip'

####################################################################################################

@task()
def get_last_version(ctx):
    from bs4 import BeautifulSoup
    import requests
    response = requests.get(RELEASE_URL)
    assert(response.status_code == requests.codes.ok)
    soup = BeautifulSoup(response.text, 'html.parser')
    divs = soup.find_all('tr', attrs={'class': 'folder'})
    for div in divs:
        if 'title' in div.attrs:
            version = div.attrs['title']
            try:
                version = int(version)
                if not hasattr(ctx, 'ngspice_last_version'):
                    ctx.ngspice_last_version = version
                date = div.find('td', attrs={'headers': 'files_date_h'}).get_text()
                print('version {} on {}'.format(version, date))
            except:
                # raise NameError('Bad version {}'.format(version))
                pass

####################################################################################################

@task(get_last_version)
def get_last_release_note(ctx):
    import requests
    url = RELEASE_NOTE_URL.format(ctx.ngspice_last_version)
    print('Get {} ...'.format(url))
    response = requests.get(url, allow_redirects=True)
    assert(response.status_code == requests.codes.ok)
    print(response.text)

####################################################################################################

def donwload_file(url, dst_path):
    import requests
    print('Get {} ... -> {}'.format(url, dst_path))
    response = requests.get(url, allow_redirects=True)
    assert(response.status_code == requests.codes.ok)
    with open(dst_path, mode='wb') as fh:
        fh.write(response.content)

####################################################################################################

def init(ctx):

    if hasattr(ctx, 'ctx.ngspice_base_path'):
        return

    ctx.ngspice_base_path = PYSPICE_SOURCE_PATH.joinpath('ngspice-{}'.format(ctx.ngspice_last_version))

    ctx.ngspice_source_path = Path(str(ctx.ngspice_base_path) + '-src')
    print('ngspice source path', ctx.ngspice_source_path)

    ctx.ngspice_build_path = Path(str(ctx.ngspice_base_path) + '-build')
    print('ngspice source build', ctx.ngspice_build_path)

    # ctx.install_path = Path('/usr', 'local', 'stow', 'ngspice-{}'.format(ctx.ngspice.version))
    ctx.install_path = ctx.ngspice_base_path
    print('ngspice install path', ctx.install_path)

####################################################################################################

def remove_directories(ctx):
    init(ctx)
    for path in (
            ctx.ngspice_source_path,
            ctx.ngspice_build_path,
            ctx.install_path,
    ):
        rc = input('remove {} ? [n]/y '.format(path))
        if rc == 'y':
            shutil.rmtree(path, ignore_errors=True)

####################################################################################################

@task(get_last_version)
def get_source(ctx, extract=True):
    init(ctx)
    if ctx.ngspice_source_path.exists():
        return
    # remove_directories(ctx)
    url = TAR_URL.format(ctx.ngspice_last_version)
    dst_path = 'ngspice-{}.tar.gz'.format(ctx.ngspice_last_version)
    donwload_file(url, dst_path)
    if extract:
        import tarfile
        tar_file = tarfile.open(dst_path)
        tar_file.extractall()
        ctx.ngspice_base_path.rename(ctx.ngspice_source_path)

####################################################################################################

@task
def configure(ctx):
    init(ctx)
    configure_path = ctx.ngspice_source_path.joinpath('configure')
    command = [
        str(configure_path),
        '--prefix={}'.format(ctx.install_path),
	'--with-ngshared',
	'--enable-xspice',
	'--enable-cider',
	'--enable-openmp',
    	'--disable-debug',
    ]
    if not ctx.ngspice_build_path.exists():
        os.mkdir(ctx.ngspice_build_path)
    with ctx.cd(str(ctx.ngspice_build_path)):
        ctx.run(' '.join(command))

####################################################################################################

@task
def build(ctx):
    init(ctx)
    with ctx.cd(str(ctx.ngspice_build_path)):
        ctx.run('make -j4')

####################################################################################################

@task
def clean(ctx):
    init(ctx)
    with ctx.cd(str(ctx.ngspice_build_path)):
        ctx.run('make clean')

####################################################################################################

@task(get_source, configure, clean, build)
def install(ctx):
    """Compile and install Ngspice library in the PySpice source, e.g. pyspice-source/ngspice-34

    Run a test using:

        LD_LIBRARY_PATH=$PWD/ngspice-34/lib:$LD_LIBRARY_PATH ./bin/pyspice-post-installation --check-install
    """
    if sys.platform != 'linux':
        return
    init(ctx)
    with ctx.cd(str(ctx.ngspice_build_path)):
        ctx.run('make install')

####################################################################################################

@task(get_last_version)
def get_manual(ctx):
    url = MANUAL_URL.format(ctx.ngspice_last_version)
    dst_path = 'ngspice-manual-{}.pdf'.format(ctx.ngspice_last_version)
    donwload_file(url, dst_path)

####################################################################################################

# @task(get_last_version)
# def get_osx(ctx):
#     url = OSX_URL.format(ctx.ngspice_last_version)
#     dst_path = 'ngspice-{}.pkg'.format(ctx.ngspice_last_version)
#     donwload_file(url, dst_path)

####################################################################################################

@task(get_last_version)
def get_windows(ctx):
    url = WINDOWS_URL.format(ctx.ngspice_last_version)
    dst_path = 'ngspice-{}_64.zip'.format(ctx.ngspice_last_version)
    donwload_file(url, dst_path)

####################################################################################################

@task(get_last_version)
def get_windows_dll(ctx):
    # version = ctx.ngspice_last_version
    version = NgSpice.NGSPICE_SUPPORTED_VERSION
    url = WINDOWS_DLL_URL.format(version)
    dst_path = 'ngspice-{}_dll_64.zip'.format(version)
    donwload_file(url, dst_path)



####################################################################################################

from PySpice.Spice.Netlist import SubCircuitFactory
from PySpice.Unit import *

####################################################################################################

class RingModulator(SubCircuitFactory):

    NAME = 'RingModulator'
    NODES = ('input_plus', 'input_minus',
             'carrier_plus', 'carrier_minus',
             'output_plus', 'output_minus')

    ##############################################

    def __init__(self,
                 outer_inductance,
                 inner_inductance,
                 coupling,
                 diode_model,
                ):

        super().__init__()

        input_inductor = self.L('input', 'input_plus', 'input_minus', outer_inductance)
        top_inductor = self.L('input_top', 'input_top', 'carrier_plus', inner_inductance)
        bottom_inductor = self.L('input_bottom', 'carrier_plus', 'input_bottom', inner_inductance)
        self.CoupledInductor('input_top', input_inductor.name, top_inductor.name, coupling)
        self.CoupledInductor('input_bottom', input_inductor.name, bottom_inductor.name, coupling)

        self.X('D1', diode_model, 'input_top', 'output_top')
        self.X('D2', diode_model, 'output_top', 'input_bottom')
        self.X('D3', diode_model, 'input_bottom', 'output_bottom')
        self.X('D4', diode_model, 'output_bottom', 'input_top')

        top_inductor = self.L('output_top', 'output_top', 'carrier_minus', inner_inductance)
        bottom_inductor = self.L('output_bottom', 'carrier_minus', 'output_bottom', inner_inductance)
        output_inductor = self.L('output', 'output_plus', 'output_minus', outer_inductance)
        self.CoupledInductor('output_top', output_inductor.name, top_inductor.name, coupling)
        self.CoupledInductor('output_bottom', output_inductor.name, bottom_inductor.name, coupling)



###################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

"""This modules implements classes to perform simulations.
"""

####################################################################################################

import logging
import os

####################################################################################################

from ..Config import ConfigInstall
from ..Tools.StringTools import join_list, join_dict, str_spice
from ..Unit import Unit, as_V, as_A, as_s, as_Hz, as_Degree, u_Degree

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class AnalysisParameters:

    """Base class for analysis parameters"""

    ANALYSIS_NAME = None

    ##############################################

    @property
    def analysis_name(self):
        return self.ANALYSIS_NAME

    ##############################################

    def to_list(self):
        return ()

    ##############################################

    def __str__(self):
        return '.{0.analysis_name} {1}'.format(self, join_list(self.to_list()))

####################################################################################################

class OperatingPointAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for operating point analysis."""

    ANALYSIS_NAME = 'op'

####################################################################################################

class DcSensitivityAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for DC sensitivity analysis."""

    ANALYSIS_NAME = 'sens'

    ##############################################

    def __init__(self, output_variable):
        self._output_variable = output_variable

    ##############################################

    @property
    def output_variable(self):
        return self._output_variable

    ##############################################

    def to_list(self):
        return (self._output_variable,)

####################################################################################################

class AcSensitivityAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for AC sensitivity analysis."""

    ANALYSIS_NAME = 'sens'

    ##############################################

    def __init__(self, output_variable, variation, number_of_points, start_frequency, stop_frequency):

        if variation not in ('dec', 'oct', 'lin'):
            raise ValueError("Incorrect variation type")

        self._output_variable = output_variable
        self._variation = variation
        self._number_of_points = number_of_points
        self._start_frequency = as_Hz(start_frequency)
        self._stop_frequency = as_Hz(stop_frequency)

    ##############################################

    @property
    def output_variable(self):
        return self._output_variable

    @property
    def variation(self):
        return self._variation

    @property
    def number_of_points(self):
        return self._number_of_points

    @property
    def start_frequency(self):
        return self._start_frequency

    @property
    def stop_frequency(self):
        return self._stop_frequency

    ##############################################

    def to_list(self):
        return (
            self._output_variable,
            self._variation,
            self._number_of_points,
            self._start_frequency,
            self._stop_frequency
        )

####################################################################################################

class DCAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for DC analysis."""

    ANALYSIS_NAME = 'dc'

    ##############################################

    def __init__(self, **kwargs):

        self._parameters = []
        for variable, value_slice in kwargs.items():
            variable_lower = variable.lower()
            if variable_lower[0] in ('v', 'i', 'r') or variable_lower == 'temp':
                self._parameters += [variable, value_slice.start, value_slice.stop, value_slice.step]
            else:
                raise NameError('Sweep variable must be a voltage/current source, '
                                'a resistor or the circuit temperature')

    ##############################################

    @property
    def parameters(self):
        return self._parameters

    ##############################################

    def to_list(self):
        return self._parameters

####################################################################################################

class ACAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for AC analysis."""

    ANALYSIS_NAME = 'ac'

    ##############################################

    def __init__(self, variation, number_of_points, start_frequency, stop_frequency):

        # Fixme: use mixin

        if variation not in ('dec', 'oct', 'lin'):
            raise ValueError("Incorrect variation type")

        self._variation = variation
        self._number_of_points = number_of_points
        self._start_frequency = as_Hz(start_frequency)
        self._stop_frequency = as_Hz(stop_frequency)

    ##############################################

    @property
    def variation(self):
        return self._variation

    @property
    def number_of_points(self):
        return self._number_of_points

    @property
    def start_frequency(self):
        return self._start_frequency

    @property
    def stop_frequency(self):
        return self._stop_frequency

    ##############################################

    def to_list(self):
        return (
            self._variation,
            self._number_of_points,
            self._start_frequency,
            self._stop_frequency
        )

####################################################################################################

class TransientAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for transient analysis."""

    ANALYSIS_NAME = 'tran'

    ##############################################

    def __init__(self, step_time, end_time, start_time=0, max_time=None, use_initial_condition=False):

        self._step_time = as_s(step_time)
        self._end_time = as_s(end_time)
        self._start_time = as_s(start_time)
        self._max_time = as_s(max_time, none=True)
        self._use_initial_condition = use_initial_condition

    ##############################################

    @property
    def step_time(self):
        return self._step_time

    @property
    def end_time(self):
        return self._end_time

    @property
    def start_time(self):
        return self._start_time

    @property
    def max_time(self):
        return self._max_time

    @property
    def use_initial_condition(self):
        return self._use_initial_condition

    ##############################################

    def to_list(self):
        return (
            self._step_time,
            self._end_time,
            self._start_time,
            self._max_time,
            'uic' if self._use_initial_condition else None,
        )

####################################################################################################

class MeasureParameters(AnalysisParameters):

    """This class defines measurements on analysis.

    """

    ANALYSIS_NAME = 'meas'

    ##############################################

    def __init__(self, analysis_type, name, *args):

        _analysis_type = str(analysis_type).upper()
        if _analysis_type not in ('AC', 'DC', 'OP', 'TRAN', 'TF', 'NOISE'):
            raise ValueError('Incorrect analysis type {}'.format(analysis_type))

        self._parameters = [_analysis_type, name, *args]

    ##############################################

    @property
    def parameters(self):
        return self._parameters

    ##############################################

    def to_list(self):
        return self._parameters

####################################################################################################

class PoleZeroAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for pole-zero analysis."""

    ANALYSIS_NAME = 'pz'

    ##############################################

    def __init__(self, node1, node2, node3, node4, tf_type, pz_type):

        self._nodes = (node1, node2, node3, node4)
        self._tf_type = tf_type   # transfert_function
        self._pz_type = pz_type   # pole_zero

    ##############################################

    @property
    def node1(self):
        return self._nodes[0]

    @property
    def node2(self):
        return self._nodes[1]

    def node3(self):
        return self._nodes[2]

    @property
    def node4(self):
        return self._nodes[3]

    @property
    def tf_type(self):
        return self._tf_type

    @property
    def pz_type(self):
        return self._pz_type

    ##############################################

    def to_list(self):
        return list(self._nodes) + [self._tf_type, self._pz_type]

####################################################################################################

class NoiseAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for noise analysis."""

    ANALYSIS_NAME = 'noise'

    ##############################################

    def __init__(self, output, src, variation, points, start_frequency, stop_frequency, points_per_summary):

        self._output = output
        self._src = src
        self._variation = variation
        self._points = points
        self._start_frequency = start_frequency
        self._stop_frequency = stop_frequency
        self._points_per_summary = points_per_summary

    ##############################################

    @property
    def output(self):
        return self._output

    @property
    def src(self):
        return self._src

    @property
    def variation(self):
        return self._variation

    @property
    def points(self):
        return self._points

    # Fixme: mixin
    @property
    def start_frequency(self):
        return self._start_frequency

    @property
    def stop_frequency(self):
        return self._stop_frequency

    @property
    def points_per_summary(self):
        return self._points_per_summary

    ##############################################

    def to_list(self):

        parameters = [
            self._output,
            self._src,
            self._variation,
            self._points,
            self._start_frequency,
            self._stop_frequency,
        ]

        if self._points_per_summary:
            parameters.append(self._points_per_summary)

        return parameters

####################################################################################################

class DistortionAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for distortion analysis."""

    ANALYSIS_NAME = 'disto'

    ##############################################

    def __init__(self, variation, points, start_frequency, stop_frequency, f2overf1):

        self._variation = variation
        self._points = points
        self._start_frequency = start_frequency
        self._stop_frequency = stop_frequency
        self._f2overf1 = f2overf1

    ##############################################

    @property
    def variation(self):
        return self._variation

    @property
    def points(self):
        return self._points

    @property
    def start_frequency(self):
        return self._start_frequency

    @property
    def stop_frequency(self):
        return self._stop_frequency

    @property
    def f2overf1(self):
        return self._f2overf1

    ##############################################

    def to_list(self):

        parameters = [
            self._variation,
            self._points,
            self._start_frequency,
            self._stop_frequency,
        ]

        if self._f2overf1:
            parameters.append(self._f2overf1)

        return parameters

####################################################################################################

class TransferFunctionAnalysisParameters(AnalysisParameters):

    """This class defines analysis parameters for transfer function (.tf) analysis."""

    ANALYSIS_NAME = 'tf'

    ##############################################

    def __init__(self, outvar, insrc):
        self._outvar = outvar
        self._insrc = insrc

    ##############################################

    @property
    def outvar(self):
        return self._outvar

    @property
    def insrc(self):
        return self._insrc

    ##############################################

    def to_list(self):
        return (self._outvar, self._insrc)

####################################################################################################

class CircuitSimulation:

    """Define and generate the spice instruction to perform a circuit simulation.

    .. warning:: In some cases NgSpice can perform several analyses one after the other. This case
      is partially supported.

    """

    _logger = _module_logger.getChild('CircuitSimulation')

    ##############################################

    def __init__(self, circuit, **kwargs):

        self._circuit = circuit

        self._options = {}   # .options
        self._measures = []   # .measure
        self._initial_condition = {}   # .ic
        self._node_set = {}   # .nodeset
        self._saved_nodes = set()
        self._analyses = {}

        self.temperature = kwargs.get('temperature', u_Degree(27))
        self.nominal_temperature = kwargs.get('nominal_temperature', u_Degree(27))

    ##############################################

    @property
    def circuit(self):
        return self._circuit

    ##############################################

    def options(self, *args, **kwargs):
        for item in args:
            self._options[str(item)] = None
        for key, value in kwargs.items():
            self._options[str(key)] = str_spice(value)

    ##############################################

    @property
    def temperature(self):
        return self._options['TEMP']

    @temperature.setter
    def temperature(self, value):
        self._options['TEMP'] = as_Degree(value)

    ##############################################

    @property
    def nominal_temperature(self):
        return self._options['TNOM']

    @nominal_temperature.setter
    def nominal_temperature(self, value):
        self._options['TNOM'] = as_Degree(value)

    ##############################################

    @staticmethod
    def _make_initial_condition_dict(kwargs):
        return {f"V({key})": str_spice(value) for key, value in kwargs.items()}

    ##############################################

    def initial_condition(self, **kwargs):
        """Set initial condition for voltage nodes.

        Usage::

            simulator.initial_condition(node_name=value, ...)

        General form::

            .ic v(node_name)=value ...

        The `.ic` line is for setting transient initial conditions.  It has two different
        interpretations, depending on whether the uic parameter is specified on the `.tran` control
        line, or not.  One should not confuse this line with the `.nodeset` line.  The `.nodeset`
        line is only to help DC convergence, and does not affect the final bias solution (except for
        multi-stable circuits).  The two indicated interpretations of this line are as follows:

        1. When the uic parameter is specified on the `.tran` line, the node voltages specified on
           the `.ic` control line are used to compute the capacitor, diode, BJT, JFET, and MOSFET
           initial conditions.  This is equivalent to specifying the `ic=...` parameter on each
           device line, but is much more convenient.  The `ic=...` parameter can still be specified
           and takes precedence over the `.ic` values.  Since no dc bias (initial transient)
           solution is computed before the transient analysis, one should take care to specify all
           dc source voltages on the `.ic` control line if they are to be used to compute device
           initial conditions.

        2. When the uic parameter is not specified on the `.tran` control line, the DC bias (initial
           transient) solution is computed before the transient analysis. In this case, the node
           voltages specified on the `.ic` control lines are forced to the desired initial values
           during the bias solution.  During transient analysis, the constraint on these node
           voltages is removed.  This is the preferred method since it allows Ngspice to compute a
           consistent dc solution.

        """
        d = self._make_initial_condition_dict(kwargs)
        self._initial_condition.update(d)

    ##############################################

    def node_set(self, **kwargs):
        """Specify initial node voltage guesses.

        Usage::

            simulator.node_set(node_name=value, ...)

        General form::

            .nodeset v(node_name)=value ...
            .nodeset all=val

        The `.nodeset` line helps the program find the DC or initial transient solution by making a
        preliminary pass with the specified nodes held to the given voltages.  The restrictions are
        then released and the iteration continues to the true solution.  The `.nodeset` line may be
        necessary for convergence on bistable or astable circuits. `.nodeset all=val` sets all
        starting node voltages (except for the ground node) to the same value.  In general, the
        `.nodeset` line should not be necessary.

        """
        d = self._make_initial_condition_dict(kwargs)
        self._node_set.update(d)

    ##############################################

    def save(self, *args):

        # Fixme: pass Node for voltage node, Element for source branch current, ...

        """Set the list of saved vectors.

        If no *.save* line is given, then the default set of vectors is saved (node voltages and
        voltage source branch currents). If *.save* lines are given, only those vectors specified
        are saved.

        Node voltages may be saved by giving the node_name or *v(node_name)*. Currents through an
        independent voltage source (including inductor) are given by *i(source_name)* or
        *source_name#branch*. Internal device data are accepted as *@dev[param]*.

        If you want to save internal data in addition to the default vector set, add the parameter
        *all* to the additional vectors to be saved.

        """

        self._saved_nodes |= set(*args)

    ##############################################

    def save_internal_parameters(self, *args):
        """This method is similar to`save` but assume *all*."""
        # Fixme: ok ???
        self.save(list(args) + ['all'])

    ##############################################

    @property
    def save_currents(self):
        """ Save all currents. """
        return self._options.get('SAVECURRENTS', False)

    @save_currents.setter
    def save_currents(self, value):
        if value:
            self._options['SAVECURRENTS'] = True
        else:
            del self._options['SAVECURRENTS']

    ##############################################

    def reset_analysis(self):
        self._analyses.clear()

    ##############################################

    def analysis_iter(self):
        return self._analyses.values()

    ##############################################

    def _add_analysis(self, analysis_parameters):
        self._analyses[analysis_parameters.analysis_name] = analysis_parameters

    ##############################################

    def _add_measure(self, measure_parameters):
        self._measures.append(measure_parameters)

    ##############################################

    def operating_point(self):
        """Compute the operating point of the circuit with capacitors open and inductors shorted."""
        self._add_analysis(OperatingPointAnalysisParameters())

    ##############################################

    def dc_sensitivity(self, output_variable):

        """Compute the sensitivity of the DC operating point of a node voltage or voltage-source
        branch current to all non-zero device parameters.

        Examples of usage::

            analysis = simulator.dc_sensitivity('v(out)')

        Spice syntax:

        .. code:: spice

            .sens outvar

        Examples:

        .. code:: spice

            .sens V(1, OUT)
            .sens I(VTEST)

        """

        self._add_analysis(DcSensitivityAnalysisParameters(output_variable))

    ##############################################

    def ac_sensitivity(self, output_variable, variation, number_of_points, start_frequency, stop_frequency):

        """Compute the sensitivity of the AC values of a node voltage or voltage-source branch
        current to all non-zero device parameters.

        Examples of usage::

            analysis = simulator.ac_sensitivity(...)

        Spice syntax:

        .. code::

            .sens outvar ac dec nd fstart fstop
            .sens outvar ac oct no fstart fstop
            .sens outvar ac lin np fstart fstop

        Spice examples:

        .. code::

            .sens V(OUT) AC DEC 10 100 100 k

        """

        self._add_analysis(
            AcSensitivityAnalysisParameters(
                output_variable,
                variation, number_of_points, start_frequency, stop_frequency
            ))

    ##############################################

    def dc(self, **kwargs):

        """Compute the DC transfer fonction of the circuit with capacitors open and inductors shorted.

        Examples of usage::

            analysis = simulator.dc(Vinput=slice(-2, 5, .01))
            analysis = simulator.dc(Ibase=slice(0, 100e-6, 10e-6))
            analysis = simulator.dc(Vcollector=slice(0, 5, .1), Ibase=slice(micro(10), micro(100), micro(10))) # broken ???

        Spice syntax:

        .. code:: spice

            .dc src_name vstart vstop vincr [ src2 start2 stop2 incr2 ]

        *src_name* is the name of an independent voltage or a current source, a resistor or the
        circuit temperature.

        *vstart*, *vstop*, and *vincr* are the starting, final, and incrementing values respectively.

        A second source (*src2*) may optionally be specified with associated sweep parameters. In
        this case, the first source is swept over its range for each value of the second source.

        Spice examples:

        .. code:: spice

            .dc VIN 0 .2 5 5.0 0.25
            .dc VDS 0 10 .5 VGS 0 5 1
            .dc VCE 0 10 .2 5 IB 0 10U 1U
            .dc RLoad 1k 2k 100
            .dc TEMP -15 75 5

        """

        self._add_analysis(DCAnalysisParameters(**kwargs))

    ##############################################

    def ac(self, variation, number_of_points, start_frequency, stop_frequency):

        # fixme: concise keyword ?

        """Perform a small-signal AC analysis of the circuit where all non-linear devices are linearized
        around their actual DC operating point.

        Examples of usage::

            analysis = simulator.ac(start_frequency=10@u_kHz, stop_frequency=1@u_GHz, number_of_points=10,  variation='dec')

        Note that in order for this analysis to be meaningful, at least one independent source must
        have been specified with an AC value. Typically it does not make much sense to specify more
        than one AC source. If you do, the result will be a superposition of all sources, thus
        difficult to interpret.

        Spice examples:

        .. code::

            .ac dec nd fstart fstop
            .ac oct no fstart fstop
            .ac lin np fstart fstop

        The parameter *variation* must be either `dec`, `oct` or `lin`.

        """

        self._add_analysis(
            ACAnalysisParameters(
                variation, number_of_points, start_frequency, stop_frequency
            ))

    ##############################################

    def measure(self, analysis_type, name, *args):

        """Add a measure in the circuit.

        Examples of usage::

            simulator.measure('TRAN', 'tdiff', 'TRIG AT=10m', 'TARG v(n1) VAL=75.0 CROSS=1')
            simulator.measure('tran', 'tdiff', 'TRIG AT=0m', f"TARG par('v(n1)-v(derate)') VAL=0 CROSS=1")

        Note: can be used with the .options AUTOSTOP to stop the simulation at Trigger.

        Spice syntax:

        .. code:: spice

            .meas tran tdiff TRIG AT=0m TARG v(n1) VAL=75.0 CROSS=1

        """

        self._add_measure(MeasureParameters(analysis_type, name, *args))

    ##############################################

    def transient(self, step_time, end_time, start_time=0, max_time=None, use_initial_condition=False):

        """Perform a transient analysis of the circuit.

        Examples of usage::

            analysis = simulator.transient(step_time=1@u_us, end_time=500@u_us)
            analysis = simulator.transient(step_time=source.period/200, end_time=source.period*2)

        Spice syntax:

        .. code:: spice

            .tran tstep tstop <tstart <tmax>> <uic>

        """

        self._add_analysis(
            TransientAnalysisParameters(
                step_time, end_time, start_time, max_time,
                use_initial_condition
            ))

    ##############################################

    def polezero(self, node1, node2, node3, node4, tf_type, pz_type):

        """Perform a Pole-Zero analysis of the circuit.

        node1, node2 - Input node pair.
        node3, node4 - Output node pair
        tf_type - should be `cur` for current or `vol` for voltage
        pz_type - should be `pol` for pole, `zer` for zero, or `pz` for combined pole zero analysis.

        See section 15.3.6 of ngspice manual.

        Spice syntax:

        .. code:: spice

            .tran tstep tstop <tstart <tmax>> <uic>
            .pz node1 node2 node3 node4 cur pol
            .pz node1 node2 node3 node4 cur zer
            .pz node1 node2 node3 node4 cur pz
            .pz node1 node2 node3 node4 vol pol
            .pz node1 node2 NODE3 node4 vol zer
            .pz node1 node2 node3 node4 vol pz

        Examples:

        .. code:: spice

            .pz 1 0 3 0 cur pol
            .pz 2 3 5 0 vol zer
            .pz 4 1 4 1 cur pz

        """

        # do some rudimentary parameter checking.
        if tf_type not in ('cur', 'vol'):
            raise NameError("polezero type must be 'cur' or 'vol'")
        if pz_type not in ('pol', 'zer', 'pz'):
            raise NameError("pz_type must be 'pol' or 'zer' or 'pz'")

        self._add_analysis(
            PoleZeroAnalysisParameters(node1, node2, node3, node4, tf_type, pz_type)
        )

    ##############################################

    def noise(self, output_node, ref_node, src, variation, points, start_frequency, stop_frequency, points_per_summary=None):

        """Perform a Pole-Zero analysis of the circuit.

        output_node, ref_node - output node pair.
        src - signal source, typically an ac voltage input.
        variation - must be 'dec' or 'lin' or 'oct' for decade, linear, or octave.
        points, start_frequency, stop_frequency - number of points, start and stop frequencies.
        points_per_summary - if specified, the noise contributions of each noise generator is produced every points_per_summary frequency points.

        See section 15.3.4 of ngspice manual.

        Spice syntax:

        General form:

        .. code:: spice

            .noise v(output <,ref >) src ( dec | lin | oct ) pts fstart fstop <pts_per_summary >

        Examples:

        .. code:: spice

            .noise v(5) VIN dec 10 1kHz 100 MEG
            .noise v(5 ,3) V1 oct 8 1.0 1.0 e6 1

        """

        # do some rudimentary parameter checking.
        # Fixme: mixin
        if variation not in ('dec', 'lin', 'oct'):
            raise NameError("variation must be 'dec' or 'lin' or 'oct'")

        output = 'V({},{})'.format(output_node, ref_node)

        self._add_analysis(
            NoiseAnalysisParameters(output, src, variation, points, start_frequency, stop_frequency, points_per_summary)
        )

    ##############################################

    def transfer_function(self, outvar, insrc):

        """The python arguments to this function should be two strings, outvar and insrc.

        ngspice documentation as follows:

        General form:

        .. code:: spice

            .tf outvar insrc

        Examples:

        .. code:: spice

           .tf v(5, 3) VIN
           .tf i(VLOAD) VIN

        The .tf line defines the small-signal output and input for the dc small-signal
        analysis. outvar is the small signal output variable and insrc is the small-signal input
        source. If this line is included, ngspice computes the dc small-signal value of the transfer
        function (output/input), input resistance, and output resistance. For the first example,
        ngspice would compute the ratio of V(5, 3) to VIN, the small-signal input resistance at VIN,
        and the small signal output resistance measured across nodes 5 and 3

        """

        self._add_analysis(
            TransferFunctionAnalysisParameters(outvar, insrc)
        )

    ##############################################

    def distortion(self, variation, points, start_frequency, stop_frequency, f2overf1=None):

        """Perform a distortion analysis of the circuit.

        variation, points, start_frequency, stop_frequency - typical ac range parameters.
        if f2overf1 is specified, perform a spectral analysis, else perform a harmonic analysis.

        See section 15.3.3 of ngspice manual.

        - harmonic analysis,
          The distof1 parameter of the AC input to the circuit must be specified.
          Second harmonic magnitude and phase are calculated at each circuit node.

        - Spectral analysis,
           The distof2 parameter of the AC input to the circuit must be specified as well as distof1.
           See the ngspice manual.

        Spice syntax:

        General form:

        .. code:: spice

              .disto dec nd fstart fstop <f2overf1 >
              .disto oct no fstart fstop <f2overf1 >
              .disto lin np fstart fstop <f2overf1 >

        Examples:

        .. code:: spice

              .disto dec 10 1kHz 100 MEG
              .disto dec 10 1kHz 100 MEG 0.9

        """

        # do some rudimentary parameter checking.
        if variation not in ('dec', 'lin', 'oct'):
            raise NameError("variation must be 'dec' or 'lin' or 'oct'")

        self._add_analysis(
            DistortionAnalysisParameters(variation, points, start_frequency, stop_frequency, f2overf1)
        )

    ##############################################

    def str_options(self, unit=True):

        # Fixme: use cls settings ???
        if unit:
            _str = str_spice
        else:
            _str = lambda x: str_spice(x, unit)

        netlist = ''
        if self.options:
            for key, value in self._options.items():
                if value is not None:
                    netlist += '.options {} = {}'.format(key, _str(value)) + os.linesep
                else:
                    netlist += '.options {}'.format(key) + os.linesep
        return netlist

    ##############################################

    def __str__(self):

        netlist = self._circuit.str(simulator=self.SIMULATOR)
        netlist += self.str_options()
        if self._initial_condition:
            netlist += '.ic ' + join_dict(self._initial_condition) + os.linesep
        if self._node_set:
            netlist += '.nodeset ' + join_dict(self._node_set) + os.linesep

        if self._saved_nodes:
            # Place 'all' first
            saved_nodes = self._saved_nodes
            if 'all' in saved_nodes:
                all_str = 'all '
                saved_nodes.remove('all')
            else:
                all_str = ''
            netlist += '.save ' + all_str + join_list(saved_nodes) + os.linesep
        for measure_parameters in self._measures:
            netlist += str(measure_parameters) + os.linesep
        for analysis_parameters in self._analyses.values():
            netlist += str(analysis_parameters) + os.linesep
        netlist += '.end' + os.linesep
        return netlist

####################################################################################################

class CircuitSimulator(CircuitSimulation):

    """ This class implements a circuit simulator. Each analysis mode is performed by a method that
    return the measured probes.

    For *ac* and *transient* analyses, the user must specify a list of nodes using the *probes* key
    argument.
    """

    _logger = _module_logger.getChild('CircuitSimulator')

    if ConfigInstall.OS.on_windows:
        DEFAULT_SIMULATOR = 'ngspice-shared'
    else:
        # DEFAULT_SIMULATOR = 'ngspice-subprocess'
        DEFAULT_SIMULATOR = 'ngspice-shared'
        # DEFAULT_SIMULATOR = 'xyce-serial'
        # DEFAULT_SIMULATOR = 'xyce-parallel'

    ##############################################

    @classmethod
    def factory(cls, circuit, *args, **kwargs):

        """Return a :obj:`PySpice.Spice.Simulation.SubprocessCircuitSimulator` or
        :obj:`PySpice.Spice.Simulation.NgSpiceSharedCircuitSimulator` instance depending of the
        value of the *simulator* parameter: ``subprocess`` or ``shared``, respectively. If this
        parameter is not specified then a subprocess simulator is returned.

        """

        if 'simulator' in kwargs:
            simulator = kwargs['simulator']
            del kwargs['simulator']
        else:
            simulator = cls.DEFAULT_SIMULATOR

        sub_cls = None
        if simulator in ('ngspice-subprocess', 'ngspice-shared'):
            if simulator == 'ngspice-subprocess':
                from .NgSpice.Simulation import NgSpiceSubprocessCircuitSimulator
                sub_cls = NgSpiceSubprocessCircuitSimulator
            elif simulator == 'ngspice-shared':
                from .NgSpice.Simulation import NgSpiceSharedCircuitSimulator
                sub_cls = NgSpiceSharedCircuitSimulator
        elif simulator in ('xyce-serial', 'xyce-parallel'):
            from .Xyce.Simulation import XyceCircuitSimulator
            sub_cls = XyceCircuitSimulator
            if simulator == 'xyce-parallel':
                kwargs['parallel'] = True

        if sub_cls is not None:
            return sub_cls(circuit, *args, **kwargs)
        else:
            raise ValueError('Unknown simulator type')

    ##############################################

    def _run(self, analysis_method, *args, **kwargs):

        self.reset_analysis()
        if 'probes' in kwargs:
            self.save(* kwargs.pop('probes'))

        _kwargs = dict(kwargs)
        _kwargs.pop('log_desk', None)

        method = getattr(CircuitSimulation, analysis_method)
        method(self, *args, **_kwargs)

        message = 'desk' + os.linesep + str(self)
        if kwargs.get('log_desk', False):
            self._logger.info(message)
        else:
            self._logger.debug(message)

    ##############################################

    def operating_point(self, *args, **kwargs):
        return self._run('operating_point', *args, **kwargs)

    ##############################################

    def dc(self, *args, **kwargs):
        return self._run('dc', *args, **kwargs)

    ##############################################

    def dc_sensitivity(self, *args, **kwargs):
        return self._run('dc_sensitivity', *args, **kwargs)

    ##############################################

    def ac(self, *args, **kwargs):
        return self._run('ac', *args, **kwargs)

    ##############################################

    def transient(self, *args, **kwargs):
        return self._run('transient', *args, **kwargs)

    ##############################################

    def polezero(self, *args, **kwargs):
        return self._run('polezero', *args, **kwargs)

    ##############################################

    def noise(self, *args, **kwargs):
        return self._run('noise', *args, **kwargs)

    ##############################################

    def distortion(self, *args, **kwargs):
        return self._run('distortion', *args, **kwargs)

    ##############################################

    def transfer_function(self, *args, **kwargs):
        return self._run('transfer_function', *args, **kwargs)

    tf = transfer_function   # shorcut



####################################################################################################

#r# ===================
#r#  Millman's theorem
#r# ===================
#r#
#r# Millman's theorem is a method to compute the voltage of a node in such circuits:

#f# circuit_macros('millman-theorem.m4')

#r# The voltage at node A is:
#r#
#r# .. math::
#r#
#r#     V_A = \frac{\sum \frac{V_i}{R_i}}{\sum \frac{1}{R_i}}
#r#
#r# We can extend this theorem to branches with current sources:
#r#
#r# .. math::
#r#
#r#     V_A = \frac{\sum \frac{V_i}{R_i} + \sum I_k}{\sum \frac{1}{R_i}}
#r#
#r# Note voltage sources can be null and resistances in current's branches don't change the denominator.

#f# circuit_macros('millman-theorem-with-current-source.m4')

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

import numpy as np

####################################################################################################

from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

####################################################################################################

circuit = Circuit("Millman's theorem")

number_of_branches = 3
for i in range(1, number_of_branches +1):
    circuit.V('input%u' % i, i, circuit.gnd, i@u_V)
    circuit.R(i, i, 'A', i@u_kΩ)

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.operating_point()

node_A = analysis.A
print('Node {}: {:5.2f} V'.format(str(node_A), float(node_A)))
#o#

branch_voltages = np.arange(1, number_of_branches +1)
branch_resistances = branch_voltages * float(kilo(1))
conductances = 1 / branch_resistances
voltage_A = np.sum(branch_voltages * conductances) / np.sum(conductances)
print('V(A) = {:5.2f} V'.format(voltage_A))
#o#

# with current sources
for i in range(1, number_of_branches +1):
    ii = number_of_branches + i
    circuit.I('input%u' % i, circuit.gnd, ii, 100*i@u_uA)
    circuit.R(ii, ii, 'A', i@u_kΩ)

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.operating_point()

node_A = analysis.A
print('Node {}: {:5.2f} V'.format(str(node_A), float(node_A)))
#o#

branch_currents = np.arange(1, number_of_branches +1) * float(micro(100))
voltage_A += np.sum(branch_currents) / np.sum(conductances)
print('V(A) = {:5.2f} V'.format(voltage_A))
#o#



#r# This example shows a low-pass RC Filter.

####################################################################################################

import math
import numpy as np
import matplotlib.pyplot as plt

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Plot.BodeDiagram import bode_diagram
from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

####################################################################################################

#f# circuit_macros('low-pass-rc-filter.m4')

circuit = Circuit('Low-Pass RC Filter')

circuit.SinusoidalVoltageSource('input', 'in', circuit.gnd, amplitude=1@u_V)
R1 = circuit.R(1, 'in', 'out', 1@u_kΩ)
C1 = circuit.C(1, 'out', circuit.gnd, 1@u_uF)

#r# The break frequency is given by :math:`f_c = \frac{1}{2 \pi R C}`

break_frequency = 1 / (2 * math.pi * float(R1.resistance * C1.capacitance))
print("Break frequency = {:.1f} Hz".format(break_frequency))
#o#

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.ac(start_frequency=1@u_Hz, stop_frequency=1@u_MHz, number_of_points=10,  variation='dec')
# print(analysis.out)

#r# We plot the Bode diagram.

figure, axes = plt.subplots(2, figsize=(20, 10))
plt.title("Bode Diagram of a Low-Pass RC Filter")
bode_diagram(axes=axes,
             frequency=analysis.frequency,
             gain=20*np.log10(np.absolute(analysis.out)),
             phase=np.angle(analysis.out, deg=False),
             marker='.',
             color='blue',
             linestyle='-',
)
for ax in axes:
    ax.axvline(x=break_frequency, color='red')

plt.tight_layout()
plt.show()

#f# save_figure('figure', 'low-pass-rc-filter-bode-diagram.png')



#r# This example depicts a ring modulator

#r## .. warning:: It don't simulate
#r#
#r##   doAnalyses: TRAN:  Timestep too small; time = 5.5453e-08, timestep = 1.25e-18: trouble with xring_modulator.xd2:1n4148-instance d.xring_modulator.xd2.d1

####################################################################################################

import matplotlib.pyplot as plt

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Doc.ExampleTools import find_libraries
from PySpice.Probe.Plot import plot
from PySpice.Spice.Library import SpiceLibrary
from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

####################################################################################################

libraries_path = find_libraries()
spice_library = SpiceLibrary(libraries_path)

####################################################################################################

from RingModulator import RingModulator

#f# literal_include('RingModulator.py')

####################################################################################################

circuit = Circuit('Ring Modulator')

modulator = circuit.SinusoidalVoltageSource('modulator', 'in', circuit.gnd, amplitude=1@u_V, frequency=1@u_kHz)
carrier = circuit.SinusoidalVoltageSource('carrier', 'carrier', circuit.gnd, amplitude=10@u_V, frequency=100@u_kHz)
circuit.R('in', 'in', 1, 50@u_Ω)
circuit.R('carrier', 'carrier', 2, 50@u_Ω)

circuit.include(spice_library['1N4148'])
circuit.subcircuit(RingModulator(outer_inductance=1@u_uH,
                                 inner_inductance=1@u_uH,
                                 coupling=.99,
                                 diode_model='1N4148',
                             ))
circuit.X('ring_modulator', 'RingModulator',
          1, circuit.gnd,
          2, circuit.gnd,
          'output', circuit.gnd,
         )

# outer_inductance = .01
# inner_inductance = .0025
# coupling = .9
# diode_model = '1N4148'
# input_inductor = circuit.L('input', 1, circuit.gnd, outer_inductance)
# top_inductor = circuit.L('input_top', 'input_top', 'carrier', inner_inductance)
# bottom_inductor = circuit.L('input_bottom', 'input_bottom', 'carrier', inner_inductance)
# circuit.CoupledInductor('input_top', input_inductor.name, top_inductor.name, coupling)
# circuit.CoupledInductor('input_bottom', input_inductor.name, bottom_inductor.name, coupling)
# circuit.X('D1', diode_model, 'input_top', 'output_top')
# circuit.X('D2', diode_model, 'output_top', 'input_bottom')
# circuit.X('D3', diode_model, 'input_bottom', 'output_bottom')
# circuit.X('D4', diode_model, 'output_bottom', 'input_top')
# top_inductor = circuit.L('output_top', 'output_top', circuit.gnd, inner_inductance)
# bottom_inductor = circuit.L('output_bottom', 'output_bottom', circuit.gnd, inner_inductance)
# output_inductor = circuit.L('output', 'output', circuit.gnd, outer_inductance)
# circuit.CoupledInductor('output_top', output_inductor.name, top_inductor.name, coupling)
# circuit.CoupledInductor('output_bottom', output_inductor.name, bottom_inductor.name, coupling)

circuit.R('load', 'output', circuit.gnd, 1@u_kΩ)

### simulator = circuit.simulator(temperature=25, nominal_temperature=25)
### # simulator.initial_condition(input_top=0, input_bottom=0, output_top=0, output_bottom=0)
### analysis = simulator.transient(step_time=modulator.period/1000, end_time=modulator.period)
###
### figure = plt.figure(1, (20, 10))
### plt.title('Ring Modulator')
### plt.xlabel('Time [s]')
### plt.ylabel('Voltage [V]')
### plt.grid()
### plot(analysis['Vmodulator'])
### plot(analysis['Vcarrier'])
### # plot(analysis['output'])
### plt.legend(('modulator', 'carrier', 'output'), loc=(.05,.1))

####################################################################################################

plt.show()



###################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2017 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

"""This modules implements classes to perform simulations.
"""

####################################################################################################

import logging

####################################################################################################

from ..Simulation import CircuitSimulator
from .Server import XyceServer

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class XyceCircuitSimulator(CircuitSimulator):

    _logger = _module_logger.getChild('XyceCircuitSimulator')

    SIMULATOR = 'xyce'

    ##############################################

    def __init__(self, circuit, **kwargs):

        super().__init__(circuit, **kwargs)

        xyce_command = kwargs.get('xyce_command', None)
        self._xyce_server = XyceServer(xyce_command=xyce_command)

    ##############################################

    def str_options(self):

        return super().str_options(unit=False)

    ##############################################

    def _run(self, analysis_method, *args, **kwargs):

        super()._run(analysis_method, *args, **kwargs)

        raw_file = self._xyce_server(spice_input=str(self))
        self.reset_analysis()
        raw_file.simulation = self

        # for field in raw_file.variables:
        #     print field

        return raw_file.to_analysis()



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

"""This module provides a Python interface to the Ngspice shared library described in the *ngspice
as shared library or dynamic link library* section of the Ngspice user manual.

In comparison to the subprocess interface, it provides an interaction with the simulator through
commands and callbacks and it enables the usage of external voltage and current source in the
circuit.

.. This approach corresponds to the *standard way* to make an interface to a simulator code.

.. warning:: Since we don't simulate a circuit in a fresh environment on demand, this approach is
 less safe than the subprocess approach. In case of bugs within Ngspice, we can expect some side
 effects like memory leaks or worse unexpected things.

This interface use the CFFI module to interface with the shared library. It is thus suited to run
within the Pypy interpreter which implements JIT optimisation for Python.

It can also be used to experiment parallel simulation as explained in the Ngspice user manual. But
it seems the Ngspice source code was designed with global variables which imply to use one copy of
the shared library by worker as explained in the manual.

.. warning:: This interface can strongly slow down the simulation if the input or output callbacks
  is used.  If the simulation time goes wrong for you then you need to implement the callbacks at a
  lower level than Python. You can have look to Pypy, Cython or a C extension module.

"""

####################################################################################################

# 16.7 Environmental variables
# 16.7.1 Ngspice specific variables
#
# SPICE_LIB_DIR
#   default: /usr/local/share/ngspice (Linux, CYGWIN), C:\Spice\share\ngspice (Windows)
# SPICE_EXEC_DIR
#   default: /usr/local/bin (Linux, CYGWIN), C:\Spice\bin (Windows)
# SPICE_ASCIIRAWFILE
#   default: 0
#   Format of the rawfile. 0 for binary, and 1 for ascii.
# SPICE_SCRIPTS
#   default: $SPICE_LIB_DIR/scripts
#   In this directory the spinit file will be searched.
# NGSPICE_MEAS_PRECISION
#   default: 5
#   Sets the number of digits if output values are printed by the meas(ure) command.
# SPICE_NO_DATASEG_CHECK
#   default: undefined
#   If defined, will suppress memory resource info (probably obsolete, not used on Windows
#   or where the /proc information system is available.)
# NGSPICE_INPUT_DIR
#   default: undefined
#   If defined, using a valid directory name, will add the given directory to the search path
#   when looking for input files (*.cir, *.inc, *.lib).

####################################################################################################

__all__ = [
    'NgSpiceCircuitError',
    'NgSpiceCommandError',
    'NgSpiceShared',
]

####################################################################################################

from pathlib import Path
import logging
import os
import platform
import re

import numpy as np

from cffi import FFI

####################################################################################################

from PySpice.Config import ConfigInstall
from PySpice.Probe.WaveForm import (
    OperatingPoint, SensitivityAnalysis,
    DcAnalysis, AcAnalysis, TransientAnalysis,
    PoleZeroAnalysis, NoiseAnalysis, DistortionAnalysis, TransferFunctionAnalysis,
    WaveForm,
)
from PySpice.Tools.EnumFactory import EnumFactory
from PySpice.Unit import u_V, u_A, u_s, u_Hz, u_F, u_Degree

from .SimulationType import SIMULATION_TYPE

####################################################################################################

ffi = FFI()

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class NgSpiceCircuitError(NameError):
    pass

class NgSpiceCommandError(NameError):
    pass

####################################################################################################

def ffi_string_utf8(_):
    _ = ffi.string(_)
    try:
        return _.decode('utf8')
    except UnicodeDecodeError:
        return _

####################################################################################################

class Vector:

    """ This class implements a vector in a simulation output.

    Public Attributes:

      :attr:`data`

      :attr:`name`

    """

    _logger = _module_logger.getChild('Vector')

    ##############################################

    def __init__(self, ngspice_shared, name, type_, data):

        self._ngspice_shared = ngspice_shared
        self._name = str(name)
        self._type = type_
        self._data = data
        self._unit = ngspice_shared.type_to_unit(type_)
        if self._unit is None:
            self._logger.warning('Unit is None for {0._name} {0._type}'.format(self))

    ##############################################

    def __repr__(self):
        return 'variable: {0._name} {0._type}'.format(self)

    ##############################################

    @property
    def is_interval_parameter(self):
        return self._name.startswith('@')

    ##############################################

    @property
    def is_voltage_node(self):
        return self._type == self._ngspice_shared.simulation_type.voltage and not self.is_interval_parameter

    ##############################################

    @property
    def is_branch_current(self):
        return self._type == self._ngspice_shared.simulation_type.current and not self.is_interval_parameter

    ##############################################

    @property
    def simplified_name(self):

        if self.is_voltage_node and self._name.startswith('V('):
            return self._name[2:-1]
        elif self.is_branch_current:
            # return self._name.replace('#branch', '')
            return self._name[:-7]
        else:
            return self._name

    ##############################################

    def to_waveform(self, abscissa=None, to_real=False, to_float=False):

        """ Return a :obj:`PySpice.Probe.WaveForm` instance. """

        data = self._data
        if to_real:
            data = data.real
        # Fixme: else UnitValue instead of UnitValues
        # if to_float:
        #     data = float(data[0])

        if self._unit is not None:
            return WaveForm.from_unit_values(self.simplified_name, self._unit(data), abscissa=abscissa)
        else:
            return WaveForm.from_array(self.simplified_name, data, abscissa=abscissa)

####################################################################################################

class Plot(dict):

    """ This class implements a plot in a simulation output.

    Public Attributes:

      :attr:`plot_name`

    """

    ##############################################

    def __init__(self, simulation, plot_name):

        super().__init__()

        self._simulation = simulation
        self.plot_name = plot_name

    ##############################################

    def nodes(self, to_float=False, abscissa=None):
        return [variable.to_waveform(abscissa, to_float=to_float)
                for variable in self.values()
                if variable.is_voltage_node]

    ##############################################

    def branches(self, to_float=False, abscissa=None):
        return [variable.to_waveform(abscissa, to_float=to_float)
                for variable in self.values()
                if variable.is_branch_current]

    ##############################################

    def internal_parameters(self, to_float=False, abscissa=None):
        return [variable.to_waveform(abscissa, to_float=to_float)
                for variable in self.values()
                if variable.is_interval_parameter]

    ##############################################

    def elements(self, abscissa=None):
        return [variable.to_waveform(abscissa, to_float=True)
                for variable in self.values()]

    ##############################################

    def to_analysis(self):

        if self.plot_name.startswith('op'):
            return self._to_operating_point_analysis()
        elif self.plot_name.startswith('sens'):
            return self._to_sensitivity_analysis()
        elif self.plot_name.startswith('dc'):
            return self._to_dc_analysis()
        elif self.plot_name.startswith('ac'):
            return self._to_ac_analysis()
        elif self.plot_name.startswith('tran'):
            return self._to_transient_analysis()
        elif self.plot_name.startswith('disto'):
            return self._to_distortion_analysis()
        elif self.plot_name.startswith('noise'):
            return self._to_noise_analysis()
        elif self.plot_name.startswith('pz'):
            return self._to_polezero_analysis()
        elif self.plot_name.startswith('tf'):
            return self._to_transfer_function_analysis()
        else:
            raise NotImplementedError("Unsupported plot name {}".format(self.plot_name))

    ##############################################

    def _to_operating_point_analysis(self):
        return OperatingPoint(
            simulation=self._simulation,
            nodes=self.nodes(to_float=True),
            branches=self.branches(to_float=True),
            internal_parameters=self.internal_parameters(),
        )

    ##############################################

    def _to_sensitivity_analysis(self):
        # Fixme: separate v(vinput), analysis.R2.m
        return SensitivityAnalysis(
            simulation=self._simulation,
            elements=self.elements(),  # Fixme: internal parameters ???
            internal_parameters=self.internal_parameters(),
        )

    ##############################################

    def _to_dc_analysis(self):
        for name in ('v-sweep', 'i-sweep', 'temp-sweep'):
            if name in self:
                sweep_variable = self[name]
                break
        else:
            raise NotImplementedError(str(self))
        sweep = sweep_variable.to_waveform()
        return DcAnalysis(
            simulation=self._simulation,
            sweep=sweep,
            nodes=self.nodes(),
            branches=self.branches(),
            internal_parameters=self.internal_parameters(),
        )

    ##############################################

    def _to_ac_analysis(self):
        frequency = self['frequency'].to_waveform(to_real=True)
        return AcAnalysis(
            simulation=self._simulation,
            frequency=frequency,
            nodes=self.nodes(),
            branches=self.branches(),
            internal_parameters=self.internal_parameters(),
        )

    ##############################################

    def _to_transient_analysis(self):

        time = self['time'].to_waveform(to_real=True)
        return TransientAnalysis(
            simulation=self._simulation,
            time=time,
            nodes=self.nodes(abscissa=time),
            branches=self.branches(abscissa=time),
            internal_parameters=self.internal_parameters(abscissa=time),
        )

    ##############################################

    def _to_polezero_analysis(self):
        return PoleZeroAnalysis(
            simulation=self._simulation,
            nodes=self.nodes(),
            branches=self.branches(),
            internal_parameters=self.internal_parameters(),
        )

    ##############################################

    def _to_noise_analysis(self):
        return NoiseAnalysis(
            simulation=self._simulation,
            nodes=self.nodes(),
            branches=self.branches(),
            internal_parameters=self.internal_parameters(),
        )

    ##############################################

    def _to_distortion_analysis(self):
        frequency = self['frequency'].to_waveform(to_real=True)
        return DistortionAnalysis(
            simulation=self._simulation,
            frequency=frequency,
            nodes=self.nodes(),
            branches=self.branches(),
            internal_parameters=self.internal_parameters(),
        )

    ##############################################

    def _to_transfer_function_analysis(self):
        return TransferFunctionAnalysis(
            simulation=self._simulation,
            nodes=self.nodes(),
            branches=self.branches(),
            internal_parameters=self.internal_parameters(),
        )

####################################################################################################

class NgSpiceShared:

    _logger = _module_logger.getChild('NgSpiceShared')

    NGSPICE_PATH = None
    LIBRARY_PATH = None

    MAX_COMMAND_LENGTH = 1023
    NUMBER_OF_EXEC_CALLS_TO_RELEASE_MEMORY = 10_000

    ##############################################

    @classmethod
    def setup_platform(cls):

        if ConfigInstall.OS.on_windows:
            if platform.architecture()[0] != '64bit':
                raise NameError('Windows 32bit is no longer supported by NgSpice')

        _ = os.environ.get('NGSPICE_LIBRARY_PATH', None)
        if _ is not None:
            cls.LIBRARY_PATH = _
        else:
            if ConfigInstall.OS.on_windows:
                ngspice_path = Path(__file__).parent.joinpath('Spice64_dll')
                cls.NGSPICE_PATH = ngspice_path
                # path = ngspice_path.joinpath('dll-vs', 'ngspice-{version}{id}.dll')
                path = ngspice_path.joinpath('dll-vs', 'ngspice{}.dll')

            elif ConfigInstall.OS.on_osx:
                path = 'libngspice{}.dylib'

            elif ConfigInstall.OS.on_linux:
                path = 'libngspice{}.so'

            else:
                raise NotImplementedError

            cls.LIBRARY_PATH = str(path)

    ##############################################

    _instances = {}

    @classmethod
    def new_instance(cls, ngspice_id=0, send_data=False, verbose=False):
        """Create a NgSpiceShared instance"""

        # Fixme: send_data

        if ngspice_id in cls._instances:
            return cls._instances[ngspice_id]
        else:
            cls._logger.debug("New instance for id {}".format(ngspice_id))
            instance = cls(ngspice_id=ngspice_id, send_data=send_data, verbose=verbose)
            cls._instances[ngspice_id] = instance
            return instance

    ##############################################

    def __init__(self, ngspice_id=0, send_data=False, verbose=False):

        """ Set the *send_data* flag if you want to enable the output callback.

        Set the *ngspice_id* to an integer value if you want to run NgSpice in parallel.
        """

        self._ngspice_id = ngspice_id

        self._spinit_not_found = False

        self._number_of_exec_calls = 0

        self._stdout = []
        self._stderr = []
        self._error_in_stdout = None
        self._error_in_stderr = None

        self._has_cider = None
        self._has_xspice = None
        self._ngspice_version = None
        self._extensions = []

        self._library_path = None
        self._load_library(verbose)
        self._init_ngspice(send_data)

        self._is_running = False

    ##############################################

    @property
    def spinit_not_found(self):
        return self._spinit_not_found

    ##############################################

    @property
    def library_path(self):
        if self._library_path is None:
            if not self._ngspice_id:
                library_prefix = ''
            else:
                library_prefix = '{}'.format(self._ngspice_id)  # id =
            library_path = self.LIBRARY_PATH.format(library_prefix)
            self._library_path = library_path
        return self._library_path

    ##############################################

    def _load_library(self, verbose):

        if ConfigInstall.OS.on_windows:
            # https://sourceforge.net/p/ngspice/discussion/133842/thread/1cece652/#4e32/5ab8/9027
            # When environment variable SPICE_LIB_DIR is empty, ngspice looks in C:\Spice64\share\ngspice\scripts
            # Else it tries %SPICE_LIB_DIR%\scripts\spinit
            if 'SPICE_LIB_DIR' not in os.environ:
                _ = str(Path(self.NGSPICE_PATH).joinpath('share', 'ngspice'))
                os.environ['SPICE_LIB_DIR'] = _
                # self._logger.warning('Set SPICE_LIB_DIR = %s', _)

        # Fixme: not compatible with supra
        # if 'CONDA_PREFIX' in os.environ:
        #     _ = str(Path(os.environ['CONDA_PREFIX']).joinpath('share', 'ngspice'))
        #     os.environ['SPICE_LIB_DIR'] = _
        #     self._logger.warning('Set SPICE_LIB_DIR = %s', _)

        # https://sourceforge.net/p/ngspice/bugs/490
        # ngspice and Kicad do setlocale(LC_NUMERIC, "C");
        if ConfigInstall.OS.on_windows:
            self._logger.debug('locale LC_NUMERIC is not forced to C')
        elif ConfigInstall.OS.on_linux or ConfigInstall.OS.on_osx:
            self._logger.debug('Set locale LC_NUMERIC to C')
            import locale
            locale.setlocale(locale.LC_NUMERIC, 'C')

        api_path = Path(__file__).parent.joinpath('api.h')
        with open(api_path) as fh:
            ffi.cdef(fh.read())

        message = 'Load library {}'.format(self.library_path)
        self._logger.debug(message)
        if verbose:
            print(message)
        self._ngspice_shared = ffi.dlopen(self.library_path)

        # Note: cannot yet execute command

    ##############################################

    def _init_ngspice(self, send_data):

        # Ngspice API: ngSpice_Init ngSpice_Init_Sync

        self._send_char_c = ffi.callback('int (char *, int, void *)', self._send_char)
        self._send_stat_c = ffi.callback('int (char *, int, void *)', self._send_stat)
        self._exit_c = ffi.callback('int (int, bool, bool, int, void *)', self._exit)
        self._send_init_data_c = ffi.callback('int (pvecinfoall, int, void *)', self._send_init_data)
        self._background_thread_running_c = ffi.callback('int (bool, int, void *)', self._background_thread_running)

        if send_data:
            self._send_data_c = ffi.callback('int (pvecvaluesall, int, int, void *)', self._send_data)
        else:
            self._send_data_c = FFI.NULL

        self._get_vsrc_data_c = ffi.callback('int (double *, double, char *, int, void *)', self._get_vsrc_data)
        self._get_isrc_data_c = ffi.callback('int (double *, double, char *, int, void *)', self._get_isrc_data)

        self_c = ffi.new_handle(self)
        self._self_c = self_c  # To prevent garbage collection

        rc = self._ngspice_shared.ngSpice_Init(self._send_char_c,
                                               self._send_stat_c,
                                               self._exit_c,
                                               self._send_data_c,
                                               self._send_init_data_c,
                                               self._background_thread_running_c,
                                               self_c)
        if rc:
            raise NameError("Ngspice_Init returned {}".format(rc))

        ngspice_id_c = ffi.new('int *', self._ngspice_id)
        self._ngspice_id = ngspice_id_c  # To prevent garbage collection
        rc = self._ngspice_shared.ngSpice_Init_Sync(self._get_vsrc_data_c,
                                                    self._get_isrc_data_c,
                                                    FFI.NULL,  # GetSyncData
                                                    ngspice_id_c,
                                                    self_c)
        if rc:
            raise NameError("Ngspice_Init_Sync returned {}".format(rc))

        self._get_version()

        try:
            self._simulation_type = EnumFactory('SimulationType', SIMULATION_TYPE[self._ngspice_version])
        except KeyError:
            # See SimulationType.py
            self._simulation_type = EnumFactory('SimulationType', SIMULATION_TYPE['last'])
            self._logger.warning("Unsupported Ngspice version {}".format(self._ngspice_version))
        self._type_to_unit = {
            self._simulation_type.time: u_s,
            self._simulation_type.voltage: u_V,
            self._simulation_type.current: u_A,
            self._simulation_type.frequency: u_Hz,
            self._simulation_type.capacitance: u_F,
            self._simulation_type.temperature: u_Degree,
        }

        # Prevent paging output of commands (hangs)
        self.set('nomoremode')

    ##############################################

    @staticmethod
    def _send_char(message_c, ngspice_id, user_data):

        """Callback for sending output from stdout, stderr to caller"""

        self = ffi.from_handle(user_data)
        _module_logger.debug(str(ffi.string(message_c)))
        message = ffi_string_utf8(message_c)

        # split message in "<prefix><match = ' '><content>"
        prefix, _, content = message.partition(' ')
        if prefix == 'stderr':
            self._stderr.append(content)
            if content.startswith('Warning:'):
                func = self._logger.warning
            # elif content.startswith('Warning:'):
            else:
                self._error_in_stderr = True
                func = self._logger.error
                if content.strip() == "Note: can't find init file.":
                    self._spinit_not_found = True
                    self._logger.warning('spinit was not found')
            func(content)
        else:
            self._stdout.append(content)
            # Fixme: Ngspice writes error on stdout and stderr ...
            if 'error' in content.lower():
                self._error_in_stdout = True
            # if self._error_in_stdout:
            #     self._logger.warning(content)

        # Fixme: ???
        return self.send_char(message, ngspice_id)

    ##############################################

    @staticmethod
    def _send_stat(message, ngspice_id, user_data):
        """Callback for simulation status to caller"""
        self = ffi.from_handle(user_data)
        return self.send_stat(ffi_string_utf8(message), ngspice_id)

    ##############################################

    @staticmethod
    def _exit(exit_status, immediate_unloding, quit_exit, ngspice_id, user_data):
        """Callback for asking for a reaction after controlled exit"""
        self = ffi.from_handle(user_data)
        self._logger.debug('ngspice_id-{} exit status={} immediate_unloding={} quit_exit={}'.format(
            ngspice_id,
            exit_status,
            immediate_unloding,
            quit_exit))
        return exit_status

    ##############################################

    @staticmethod
    def _send_data(data, number_of_vectors, ngspice_id, user_data):
        """Callback to send back actual vector data"""
        self = ffi.from_handle(user_data)
        # self._logger.debug('ngspice_id-{} send_data [{}]'.format(ngspice_id, data.vecindex))
        actual_vector_values = {}
        for i in range(int(number_of_vectors)):
            actual_vector_value = data.vecsa[i]
            vector_name = ffi_string_utf8(actual_vector_value.name)
            value = complex(actual_vector_value.creal, actual_vector_value.cimag)
            actual_vector_values[vector_name] = value
            # self._logger.debug('    Vector: {} {}'.format(vector_name, value))
        return self.send_data(actual_vector_values, number_of_vectors, ngspice_id)

    ##############################################

    @staticmethod
    def _send_init_data(data, ngspice_id, user_data):
        """Callback to send back initialization vector data"""
        self = ffi.from_handle(user_data)
        # if self._logger.isEnabledFor(logging.DEBUG):
        #     self._logger.debug('ngspice_id-{} send_init_data'.format(ngspice_id))
        #     number_of_vectors = data.veccount
        #     for i in range(number_of_vectors):
        #         self._logger.debug('  Vector: ' + ffi_string_utf8(data.vecs[i].vecname))
        return self.send_init_data(data, ngspice_id)  # Fixme: should be a Python object

    ##############################################

    @staticmethod
    def _background_thread_running(is_running, ngspice_id, user_data):
        """Callback to indicate if background thread is runnin"""
        self = ffi.from_handle(user_data)
        self._logger.debug('ngspice_id-{} background_thread_running {}'.format(ngspice_id, is_running))
        self._is_running = is_running

    ##############################################

    @staticmethod
    def _get_vsrc_data(voltage, time, node, ngspice_id, user_data):
        """FFI Callback"""
        self = ffi.from_handle(user_data)
        return self.get_vsrc_data(voltage, time, ffi_string_utf8(node), ngspice_id)

    ##############################################

    @staticmethod
    def _get_isrc_data(current, time, node, ngspice_id, user_data):
        """FFI Callback"""
        self = ffi.from_handle(user_data)
        return self.get_isrc_data(current, time, ffi_string_utf8(node), ngspice_id)

    ##############################################

    def send_char(self, message, ngspice_id):
        """ Reimplement this callback in a subclass to process logging messages from the simulator. """
        # self._logger.debug('ngspice-{} send_char {}'.format(ngspice_id, message))
        return 0

    ##############################################

    def send_stat(self, message, ngspice_id):
        """ Reimplement this callback in a subclass to process statistic messages from the simulator. """
        # self._logger.debug('ngspice-{} send_stat {}'.format(ngspice_id, message))
        return 0

    ##############################################

    def send_data(self, actual_vector_values, number_of_vectors, ngspice_id):
        """ Reimplement this callback in a subclass to process the vector actual values. """
        return 0

    ##############################################

    def send_init_data(self, data, ngspice_id):
        """ Reimplement this callback in a subclass to process the initial data. """
        return 0

    ##############################################

    def get_vsrc_data(self, voltage, time, node, ngspice_id):
        """ Reimplement this callback in a subclass to provide external voltage source. """
        self._logger.debug('ngspice_id-{} get_vsrc_data @{} node {}'.format(ngspice_id, time, node))
        return 0

    ##############################################

    def get_isrc_data(self, current, time, node, ngspice_id):
        """ Reimplement this callback in a subclass to provide external current source. """
        self._logger.debug('ngspice_id-{} get_isrc_data @{} node {}'.format(ngspice_id, time, node))
        return 0

    ##############################################

    @staticmethod
    def _convert_string_array(array):
        strings = []
        i = 0
        while True:
            if array[i] == FFI.NULL:
                break
            strings.append(ffi_string_utf8(array[i]))
            i += 1
        return strings

    ##############################################

    @staticmethod
    def _to_python(value):
        try:
            return int(value)
        except ValueError:
            try:
                # Fixme: return float(value.replace(',', '.'))
                return float(value)
            except ValueError:
                return str(value)

    ##############################################

    @staticmethod
    def _lines_to_dicts(lines):
        if lines:
            values = dict(description=lines[0])
            values.update({
                parts[0]: NgSpiceShared._to_python(parts[1])
                for parts in map(str.split, lines)
            })
            return values
        else:
            raise ValueError

    ##############################################

    @property
    def is_running(self):
        return self._is_running

    ##############################################

    def clear_output(self):
        self._stdout = []
        self._stderr = []
        self._error_in_stdout = False
        self._error_in_stderr = False

    ##############################################

    @property
    def stdout(self):
        return os.linesep.join(self._stdout)

    @property
    def stderr(self):
        return os.linesep.join(self._stderr)

    ##############################################

    def exec_command(self, command, join_lines=True):

        """ Execute a command and return the output. """

        # Ngspice API: ngSpice_Command

        # Prevent memory leaks by periodically freeing ngspice history of past commands
        #   Each command sent to ngspice is stored in the control structures
        if self._number_of_exec_calls > self.NUMBER_OF_EXEC_CALLS_TO_RELEASE_MEMORY:
            # Clear the internal control structures
            self._ngspice_shared.ngSpice_Command(FFI.NULL)
            self._number_of_exec_calls = 0
        self._number_of_exec_calls += 1

        if len(command) > self.MAX_COMMAND_LENGTH:
            raise ValueError('Command must not exceed {} characters'.format(self.MAX_COMMAND_LENGTH))

        self._logger.debug('Execute command: {}'.format(command))

        self.clear_output()

        encoded_command = command.encode('ascii')
        rc = self._ngspice_shared.ngSpice_Command(encoded_command)

        if rc:  # Fixme: when not 0 ???
            raise NameError("ngSpice_Command '{}' returned {}".format(command, rc))

        if self._error_in_stdout or self._error_in_stderr:
            raise NgSpiceCommandError("Command '{}' failed".format(command))

        if join_lines:
            return self.stdout
        else:
            return self._stdout

    ##############################################

    def _get_version(self):

        self._ngspice_version = None
        self._has_xspice = False
        self._has_cider = False
        self._extensions = []

        output = self.exec_command('version -f')
        for line in output.split('\n'):
            match = re.match(r'\*\* ngspice\-(\d+)', line)
            if match is not None:
                self._ngspice_version = int(match.group(1))
            # if '** XSPICE extensions included' in line:
            if '** XSPICE' in line:
                self._has_xspice = True
                self._extensions.append('XSPICE')
            # if '** CIDER 1.b1 (CODECS simulator) included' in line:
            if 'CIDER' in line:
                self._has_cider = True
                self._extensions.append('CIDER')

        self._logger.debug(
            'Ngspice version %s with extensions: %s',
            self._ngspice_version,
            ', '.join(self._extensions),
        )

    ##############################################

    @property
    def ngspice_version(self):
        return self._ngspice_version

    @property
    def has_xspice(self):
        """Return True if libngspice was compiled with XSpice support."""
        return self._has_xspice

    @property
    def has_cider(self):
        """Return True if libngspice was compiled with CIDER support."""
        return self._has_cider

    ##############################################

    @property
    def simulation_type(self):
        return self._simulation_type

    def type_to_unit(self, vector_type):
        return self._type_to_unit.get(vector_type, None)

    ##############################################

    def _alter(self, command, device, kwargs):
        # Performance optimization: dispatch multiple alter commands jointly
        device_name = device.lower()
        commands = []
        commands_str_len = 0
        for key, value in kwargs.items():
            if isinstance(value, (list, tuple)):
                value = '[ ' + ' '.join(value) + ' ]'
            cmd = '{} {} {} = {}'.format(command, device_name, key, value)
            # performance optimization: collect multiple alter commands and
            #                           dispatch them jointly
            commands.append(cmd)
            commands_str_len += len(cmd)
            if commands_str_len + len(commands) > self.MAX_COMMAND_LENGTH:
                self.exec_command(';'.join(commands[:-1]))
                commands = commands[-1:]
                commands_str_len = len(commands[0])
        if commands:
            self.exec_command(';'.join(commands))

    ##############################################

    def alter_device(self, device, **kwargs):
        """Alter device parameters"""
        self._alter('alter', device, kwargs)

    ##############################################

    def alter_model(self, model, **kwargs):
        """Alter model parameters"""
        self._alter('altermod', model, kwargs)

    ##############################################

    def delete(self, debug_number):
        """Remove a trace or breakpoint"""
        self.exec_command('delete {}'.format(debug_number))

    ##############################################

    def destroy(self, plot_name='all'):
        """Release the memory holding the output data (the given plot or all plots) for the specified runs."""
        self.exec_command('destroy ' + plot_name)

    ##############################################

    def device_help(self, device):
        """Shows the user information about the devices available in the simulator. """
        return self.exec_command('devhelp ' + device.lower())

    ##############################################

    def save(self, vector):
        self.exec_command('save ' + vector)

    ##############################################

    def _show(self, command):
        lines = self.exec_command(command, join_lines=False)
        if lines:
            values = self._lines_to_dicts(lines)
            return values
        else:
            return ''

    ##############################################

    def show(self, device):
        return self._show('show ' + device.lower())

    ##############################################

    def showmod(self, device):
        return self._show('showmod ' + device.lower())

    ##############################################

    def source(self, file_path):
        """Read a ngspice input file"""
        self.exec_command('source ' + file_path)

    ##############################################

    def option(self, **kwargs):
        """Set any of the simulator variables."""
        for key, value in kwargs.items():
            self.exec_command('option {} = {}'.format(key, value))

    ##############################################

    def quit(self):
        self.set('noaskquit')
        return self.exec_command('quit')

    ##############################################

    def remove_circuit(self):
        """Removes the current circuit from the list of circuits sourced into ngspice."""
        self.exec_command('remcirc')

    ##############################################

    def reset(self):
        """Throw out any intermediate data in the circuit (e.g, after a breakpoint or after one or more
        analyses have been done already), and re-parse the input file. The circuit can then be
        re-run from it’s initial state, overriding the affect of any set or alter commands.

        """
        self.exec_command('reset')

    ##############################################

    def ressource_usage(self, *ressources):

        """Print resource usage statistics. If any resources are given, just print the usage of that resource.

        Most resources require that a circuit be loaded. Currently valid resources are:

        * decklineno    Number of lines in deck
        * netloadtime   Nelist loading time
        * netparsetime  Netlist parsing time
        * elapsed       The amount of time elapsed since the last rusage elapsed call.
        * faults        Number of page faults and context switches (BSD only).
        * space         Data space used.
        * time          CPU time used so far.
        * temp          Operating temperature.
        * tnom          Temperature at which device parameters were measured.
        * equations     Circuit Equations
        * time Total    Analysis Time
        * totiter       Total iterations
        * accept        Accepted time-points
        * rejected      Rejected time-points
        * loadtime      Time spent loading the circuit matrix and RHS.
        * reordertime   Matrix reordering time
        * lutime        L-U decomposition time
        * solvetime     Matrix solve time
        * trantime      Transient analysis time
        * tranpoints    Transient time-points
        * traniter      Transient iterations
        * trancuriters  Transient iterations for the last time point*
        * tranlutime    Transient L-U decomposition time
        * transolvetime Transient matrix solve time
        * everything    All of the above.
        """

        if not ressources:
            ressources = ['everything']

        command = 'rusage ' + ' '.join(ressources)
        lines = self.exec_command(command, join_lines=False)
        values = {}
        for line in lines:
            if '=' in line:
                parts = line.split(' = ')
            else:
                parts = line.split(': ')
            values[parts[0]] = NgSpiceShared._to_python(parts[1])
        return values

    ##############################################

    def set(self, *args, **kwargs):
        """Set the value of variables"""
        for key in args:
            self.exec_command('set {}'.format(key))
        for key, value in kwargs.items():
            self.exec_command('option {} = {}'.format(key, value))

    ##############################################

    def set_circuit(self, name):
        """Change the current circuit"""
        self.exec_command('setcirc {}'.format(name))

    ##############################################

    def status(self):
        """Display breakpoint information"""
        return self.exec_command('status')

    ##############################################

    def step(self, number_of_steps=None):
        """Run a fixed number of time-points"""
        if number_of_steps is not None:
            self.exec_command('step {}'.format(number_of_steps))
        else:
            self.exec_command('step')

    ##############################################

    def stop(self, *args, **kwargs):

        """Set a breakpoint.

        Examples::

            ngspice.stop('v(out) > 1', 'v(1) > 10', after=10)

        A when condition can use theses symbols: = <> > < >= <=.

        """

        command = 'stop'
        if 'after' in kwargs:
            command += ' after {}'.format(kwargs['after'])
        for condition in args:
            command += ' when {}'.format(condition)
        self.exec_command(command)

    ##############################################

    def trace(self, *args):
        """Trace nodes"""
        self.exec_command('trace ' + ' '.join(args))

    ##############################################

    def unset(self, *args):
        """Unset variables"""
        for key in args:
            self.exec_command('unset {}'.format(key))

    ##############################################

    def where(self):
        """Identify troublesome node or device"""
        return self.exec_command('where')

    ##############################################

    def load_circuit(self, circuit):

        """Load the given circuit string."""

        # Ngspice API: ngSpice_Circ
        circuit_lines = [line for line in str(circuit).splitlines() if line]
        self._logger.debug('ngSpice_Circ\n' + str(circuit))

        # ngspice 33 requires an empty line at the end
        circuit_lines.append("")

        circuit_lines_keepalive = [ffi.new("char[]", line.encode('utf8'))
                                   for line in circuit_lines]
        circuit_lines_keepalive += [FFI.NULL]
        circuit_array = ffi.new("char *[]", circuit_lines_keepalive)
        self.clear_output()
        rc = self._ngspice_shared.ngSpice_Circ(circuit_array)

        if rc:  # Fixme: when not 0 ???
            raise NameError("ngSpice_Circ returned {}".format(rc))

        # Fixme: when Ngspice found an error in the circuit, it reports the error in stdout
        # Fixme: https://sourceforge.net/p/ngspice/bugs/496/
        if self._error_in_stdout:
            self._logger.error('\n' + self.stdout)
            raise NgSpiceCircuitError('')

        # for line in circuit_lines:
        #     rc = self._ngspice_shared.ngSpice_Command(('circbyline ' + line).encode('utf8'))
        #     if rc:
        #         raise NameError("ngSpice_Command circbyline returned {}".format(rc))

    ##############################################

    def listing(self):
        command = 'listing'
        return self.exec_command(command)

    ##############################################

    def run(self, background=False):

        """ Run the simulation. """

        #  in the background thread and wait until the simulation is done

        command = 'bg_run' if background else 'run'
        self.exec_command(command)

        if background:
            self._is_running = True
        else:
            self._logger.debug("Simulation is done")

        # time.sleep(.1) # required before to test if the simulation is running
        # while (self._ngspice_shared.ngSpice_running()):
        #     time.sleep(.1)
        #     self._logger.debug("Simulation is done")

    ##############################################

    def halt(self):
        """ Halt the simulation in the background thread. """
        self.exec_command('bg_halt')

    ##############################################

    def resume(self, background=True):
        """ Halt the simulation in the background thread. """
        command = 'bg_resume' if background else 'resume'
        self.exec_command(command)

    ##############################################

    @property
    def plot_names(self):
        """ Return the list of plot names. """
        # Ngspice API: ngSpice_AllPlots
        return self._convert_string_array(self._ngspice_shared.ngSpice_AllPlots())

    ##############################################

    @property
    def last_plot(self):
        """ Return the last plot name. """
        return self.plot_names[0]

    ##############################################

    @staticmethod
    def _flags_to_str(flags):

        # enum dvec_flags {
        #   VF_REAL = (1 << 0),		// The data is real.
        #   VF_COMPLEX = (1 << 1),	// The data is complex.
        #   VF_ACCUM = (1 << 2),	// writedata should save this vector.
        #   VF_PLOT = (1 << 3),		// writedata should incrementally plot it.
        #   VF_PRINT = (1 << 4),	// writedata should print this vector.
        #   VF_MINGIVEN = (1 << 5),	// The v_minsignal value is valid.
        #   VF_MAXGIVEN = (1 << 6),	// The v_maxsignal value is valid.
        #   VF_PERMANENT = (1 << 7)	// Don't garbage collect this vector.
        # };

        if flags & 1:
            return 'real'
        elif flags & 2:
            return 'complex'
        else:
            raise NotImplementedError

    ##############################################

    @staticmethod
    def _vector_is_real(flags):
        return flags & 1

    ##############################################

    @staticmethod
    def _vector_is_complex(flags):
        return flags & 2

    ##############################################

    def plot(self, simulation, plot_name):

        """ Return the corresponding plot. """

        # Ngspice API: ngSpice_AllVecs ngGet_Vec_Info

        # plot_name is for example dc with an integer suffix which is increment for each run

        plot = Plot(simulation, plot_name)
        all_vectors_c = self._ngspice_shared.ngSpice_AllVecs(plot_name.encode('utf8'))
        i = 0
        while True:
            if all_vectors_c[i] == FFI.NULL:
                break

            vector_name = ffi_string_utf8(all_vectors_c[i])
            name = '.'.join((plot_name, vector_name))
            vector_info = self._ngspice_shared.ngGet_Vec_Info(name.encode('utf8'))
            vector_type = self._simulation_type[vector_info.v_type]
            length = vector_info.v_length
            # template = 'vector[{}] {} type {} flags {} length {}'
            # self._logger.debug(template.format(
            #     i,
            #     vector_name,
            #     vector_type,
            #     self._flags_to_str(vector_info.v_flags),
            #     length,
            # ))
            if vector_info.v_compdata == FFI.NULL:
                # for k in range(length):
                #     print("  [{}] {}".format(k, vector_info.v_realdata[k]))
                tmp_array = np.frombuffer(ffi.buffer(vector_info.v_realdata, length*8), dtype=np.float64)
                array = np.array(tmp_array, dtype=tmp_array.dtype)  # copy data
                # import json
                # with open(name + '.json', 'w') as fh:
                #     json.dump(list(array), fh)
            else:
                # for k in range(length):
                #     value = vector_info.v_compdata[k]
                #     print(ffi.addressof(value, field='cx_real'), ffi.addressof(value, field='cx_imag'))
                #     print("  [{}] {} + i {}".format(k, value.cx_real, value.cx_imag))
                tmp_array = np.frombuffer(ffi.buffer(vector_info.v_compdata, length*8*2), dtype=np.float64)
                array = np.array(tmp_array[0::2], dtype=np.complex128)
                array.imag = tmp_array[1::2]
            plot[vector_name] = Vector(self, vector_name, vector_type, array)

            i += 1

        return plot

####################################################################################################
#
# Platform setup
#

NgSpiceShared.setup_platform()



#! /usr/bin/env python3
#skip#

# Program to test pole-zero, noise, and distortion function of ngspice, PySpice

import argparse

####################################################################################################

import numpy as np
import matplotlib.pyplot as plt

####################################################################################################

import PySpice.Logging.Logging as Logging

####################################################################################################

from PySpice.Spice.Netlist import Circuit

# This program implements the PySpice equivalent of the following circuit,
# which is run using the shell command
# $ ngspice -b file.ckt
FILE_CKT = """
.title test circuit
.model qbc847b npn (BF=324.4 BR=8.29 CJC=3.347e-12 CJE=1.244e-11 CJS=0 EG=1.11 FC=0.979 IKF=0.109 IKR=0.09 IRB=5e-06 IS=1.822e-14 ISC=9.982e-12 ISE=2.894e-16 ITF=0.3131 MJC=0.391 MJE=0.3656 MJS=0.333 NC=1.763 NE=1.4 NF=0.9932 NR=0.9931 PTF=0 RB=10 RBM=5 RC=0.7014 RE=0.649 TF=4.908e-10 TR=9e-08 VAF=82 VAR=17.9 VJC=0.5463 VJE=0.7579 VJS=0.75 VTF=2.927 XCJC=0.6193 XTB=0 XTF=9.51 XTI=3)
R1 3 2 2200k
Q1 3 2 0 qbc847b
R3 1 3 1k
C1 4 2 100e-9
C2 3 5 1e-6
C4 1 3 10e-9
R5 5 0 10k
Vpwr 1 0 dc 6
Vin 4 0 dc 0 ac .001 distof1 1 distof2 0.1
.control
op 27
print all
ac dec 10 10 1e5
print V(5)
pz 4 0 5 0 vol pz
print all
noise v(5,0) Vin dec 10 10 1e5 1
print all
disto dec 10 10 1e5
print V(5)
disto dec 10 10 1e5 0.9
print V(5)
"""

class NodeNames:
    """ Allow setting of nodes with appropriate names. """
    def __init__(self, *args):
        for arg in args:
            setattr(self, arg, arg)

# npn gp signal transistor
def set_model_qbc847b(circuit):
    model = 'qbc847b'
    circuit.model(
        model, 'npn', IS=1.822E-14, NF=0.9932, ISE=2.894E-16, NE=1.4, BF=324.4,
        IKF=0.109, VAF=82, NR=0.9931, ISC=9.982E-12, NC=1.763, BR=8.29, IKR=0.09, VAR=17.9,
        RB=10, IRB=5E-06, RBM=5, RE=0.649, RC=0.7014, XTB=0, EG=1.11, XTI=3, CJE=1.244E-11,
        VJE=0.7579, MJE=0.3656, TF=4.908E-10, XTF=9.51, VTF=2.927, ITF=0.3131, PTF=0, CJC=3.347E-12,
        VJC=0.5463, MJC=0.391, XCJC=0.6193, TR=9E-08, CJS=0, VJS=0.75, MJS=0.333, FC=0.979,
    )
    return model

def simple_bjt_amp():
    circuit = Circuit('test circuit')
    model_npn = set_model_qbc847b(circuit)
    n = NodeNames('n1', 'n2', 'n3', 'n4', 'n5')
    gnd = 0
    circuit.R('1', n.n3, n.n2, 2.2e6)
    circuit.Q('1', n.n3, n.n2, gnd, model=model_npn)
    circuit.R('3', n.n1, n.n3, 1e3)
    circuit.C('1', n.n4, n.n2, 100e-9)
    circuit.C('2', n.n3, n.n5, 1e-6)
    circuit.C('4', n.n1, n.n3, 10e-9)
    circuit.R('5', n.n5, gnd, 10e3)
    circuit.V('pwr', n.n1, gnd, 6)
    circuit.V('in', n.n4, gnd, 'dc 0 ac 1 distof1 1 distof2 0.1')
    return circuit, n

def dump_circuit():
    circuit, n = simple_bjt_amp()
    print(circuit)

def do_dc_analysis():
    circuit, n = simple_bjt_amp()
    simulator = circuit.simulator(temperature=25, nominal_temperature=25)
    analysis = simulator.operating_point()
    print("Collector voltage: ", np.array(analysis[n.n3])[0])
    print("base voltage: ", np.array(analysis[n.n2])[0])

def do_ac_analysis():
    circuit, n = simple_bjt_amp()
    simulator = circuit.simulator(temperature=25, nominal_temperature=25)
    analysis = simulator.ac(start_frequency=10, stop_frequency=1e6, number_of_points=100, variation='dec')
    gain = np.array(analysis[n.n5])
    figure = plt.figure(1, (20, 10))
    axe = plt.subplot(211)
    axe.grid(True)
    axe.set_xlabel("Frequency [Hz]")
    axe.set_ylabel("dB gain.")
    axe.semilogx(analysis.frequency, 20*np.log10(np.abs(gain)))

    axe = plt.subplot(212)
    axe.grid(True)
    axe.set_xlabel("Frequency [Hz]")
    axe.set_ylabel("Phase.")
    axe.semilogx(analysis.frequency, np.arctan2(gain.imag, gain.real))
    plt.show()

def do_pz_analysis():
    circuit, n = simple_bjt_amp()
    com = 0
    simulator = circuit.simulator(temperature=25, nominal_temperature=25)
    analysis = simulator.polezero(n.n4, com, n.n5, com, 'vol', 'pz')
    print("Poles")
    for n in analysis.nodes:
        if not n.startswith('pole'):
            continue
        pole = np.array(analysis[n])
        print(pole)
    print("Zeros")
    for n in analysis.nodes:
        if not n.startswith('zero'):
            continue
        zero = np.array(analysis[n])
        print(zero)

def do_noise_analysis():
    circuit, n = simple_bjt_amp()
    com = 0
    simulator = circuit.simulator(temperature=25, nominal_temperature=25)
    analysis = simulator.noise(output_node=n.n5, ref_node=com, src='Vin', variation='dec', points=10, start_frequency=100, stop_frequency=1e5, points_per_summary=1)
    print("Total noise (Vrms) at circuit output:", np.array(analysis.nodes['onoise_total'])[0])
    print("Total noise (Vrms) as if at circuit input:", np.array(analysis.nodes['inoise_total'])[0])

def do_distortion_analysis(f2overf1):
    circuit, n = simple_bjt_amp()
    com = 0
    simulator = circuit.simulator(temperature=25, nominal_temperature=25)
    analysis = simulator.distortion(variation='dec', points=10, start_frequency=100, stop_frequency=1e4, f2overf1=f2overf1)
    output = np.array(analysis[n.n5])
    analysis_ac = simulator.ac(start_frequency=100, stop_frequency=1e4, number_of_points=10, variation='dec')
    gain = np.array(analysis_ac[n.n5])
    print("output", output)
    print("gain", gain)
    figure = plt.figure(1, (20, 10))
    axe = plt.subplot(211)
    axe.grid(True)
    axe.set_xlabel("Frequency [Hz]")
    if f2overf1:
        axe.set_ylabel("Harmonic Magnitude.")
    else:
        axe.set_ylabel("Spectral Magnitude.")
    harmonics = output/gain
    axe.semilogx(analysis.frequency, np.abs(harmonics))
    axe = plt.subplot(212)
    axe.grid(True)
    axe.set_xlabel("Frequency [Hz]")
    axe.set_ylabel("Phase.")
    axe.semilogx(analysis.frequency, np.arctan2(harmonics.imag, harmonics.real))
    plt.show()

if __name__ == '__main__':

    logger = Logging.setup_logging()

    parser = argparse.ArgumentParser("Test pyspice pole-zero, noise, distortion analysis")
    parser.add_argument('-dcir', action='store_true', dest='dump_circuit', help='dump_circuit.')
    parser.add_argument('-ac', action='store_true', dest='do_ac', help='Output AC analysis.')
    parser.add_argument('-dc', action='store_true', dest='do_dc', help='Output DC bias points.')
    parser.add_argument('-pz', action='store_true', dest='do_pz', help='Output poles and zeros.')
    parser.add_argument('-n', action='store_true', dest='do_noise', help='Output noise analysis.')
    parser.add_argument('-dh', action='store_true', dest='do_distoh', help='Output harmonic distortion analysis.')
    parser.add_argument('-ds', action='store_true', dest='do_distos', help='Output spectral distortion analysis.')

    args = parser.parse_args()

    if args.dump_circuit:
        dump_circuit()
    if args.do_ac:
        do_ac_analysis()
    if args.do_dc:
        do_dc_analysis()
    if args.do_pz:
        do_pz_analysis()
    if args.do_noise:
        do_noise_analysis()
    if args.do_distoh:
        do_distortion_analysis(None)
    if args.do_distos:
        do_distortion_analysis(0.9)



####################################################################################################

import matplotlib.pyplot as plt

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Doc.ExampleTools import find_libraries
from PySpice.Probe.Plot import plot
from PySpice.Spice.Library import SpiceLibrary
from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

####################################################################################################

libraries_path = find_libraries()
spice_library = SpiceLibrary(libraries_path)

####################################################################################################

#?# circuit_macros('buck-converter.m4')

circuit = Circuit('Buck Converter')

circuit.include(spice_library['1N5822']) # Schottky diode
circuit.include(spice_library['irf150'])

# From Microchip WebSeminars - Buck Converter Design Example

Vin = 12@u_V
Vout = 5@u_V
ratio = Vout / Vin

Iload = 2@u_A
Rload = Vout / (.8 * Iload)

frequency = 400@u_kHz
period = frequency.period
duty_cycle = ratio * period

ripple_current = .3 * Iload # typically 30 %
ripple_voltage = 50@u_mV

print('ratio =', ratio)
print('RLoad =', Rload)
print('period =', period.canonise())
print('duty_cycle =', duty_cycle.canonise())
print('ripple_current =', ripple_current)

#r# .. math:
#r#      U = L \frac{dI}{dt}

L = (Vin - Vout) * duty_cycle / ripple_current
RL = 37@u_mΩ

#r# .. math:
#r#      dV = dI (ESR + \frac{dt}{C} + \frac{ESL}{dt})

ESR = 30@u_mΩ
ESL = 0
Cout = (ripple_current * duty_cycle) / (ripple_voltage - ripple_current * ESR)

ripple_current_in = Iload / 2
ripple_voltage_in = 200@u_mV
ESR_in = 120@u_mΩ
Cin = duty_cycle / (ripple_voltage_in / ripple_current_in - ESR_in)

L = L.canonise()
Cout = Cout.canonise()
Cin = Cin.canonise()

print('L =', L)
print('Cout =', Cout)
print('Cint =', Cin)

circuit.V('in', 'in', circuit.gnd, Vin)
circuit.C('in', 'in', circuit.gnd, Cin)

# Fixme: out drop from 12V to 4V
# circuit.VCS('switch', 'gate', circuit.gnd, 'in', 'source', model='Switch', initial_state='off')
# circuit.PulseVoltageSource('pulse', 'gate', circuit.gnd, 0@u_V, Vin, duty_cycle, period)
# circuit.model('Switch', 'SW', ron=1@u_mΩ, roff=10@u_MΩ)

# Fixme: Vgate => Vout ???
circuit.X('Q', 'irf150', 'in', 'gate', 'source')
# circuit.PulseVoltageSource('pulse', 'gate', 'source', 0@u_V, Vin, duty_cycle, period)
circuit.R('gate', 'gate', 'clock', 1@u_Ω)
circuit.PulseVoltageSource('pulse', 'clock', circuit.gnd, 0@u_V, 2.*Vin, duty_cycle, period)

circuit.X('D', '1N5822', circuit.gnd, 'source')
circuit.L(1, 'source', 1, L)
circuit.R('L', 1, 'out', RL)
circuit.C(1, 'out', circuit.gnd, Cout) # , initial_condition=0@u_V
circuit.R('load', 'out', circuit.gnd, Rload)

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.transient(step_time=period/300, end_time=period*150)

figure, ax = plt.subplots(figsize=(20, 10))

ax.plot(analysis.out)
ax.plot(analysis['source'])
# ax.plot(analysis['source'] - analysis['out'])
# ax.plot(analysis['gate'])
ax.axhline(y=float(Vout), color='red')
ax.legend(('Vout [V]', 'Vsource [V]'), loc=(.8,.8))
ax.grid()
ax.set_xlabel('t [s]')
ax.set_ylabel('[V]')

plt.tight_layout()
plt.show()

#f# save_figure('figure', 'buck-converter.png')
